"""Console script for deeprenewal."""
import random
import os
import numpy as np
import mxnet as mx
from pathlib import Path
from tqdm.autonotebook import tqdm

def seed_everything():
    random.seed(42)
    np.random.seed(42)
    mx.random.seed(42)


seed_everything()

from gluonts.trainer import Trainer
from gluonts.evaluation.backtest import make_evaluation_predictions
from argparse import ArgumentParser
import sys
import ast
from deeprenewal import get_dataset, DeepRenewalEstimator, IntermittentEvaluator


def evaluate_forecast(args):
    is_gpu = mx.context.num_gpus() > 0
    dataset = get_dataset(args.datasource, regenerate=args.regenerate_datasource)
    prediction_length = dataset.metadata.prediction_length
    freq = dataset.metadata.freq
    cardinality = (
        ast.literal_eval(dataset.metadata.feat_static_cat[0].cardinality)
        if args.use_feat_static_cat
        else None
    )
    train_ds = dataset.train
    test_ds = dataset.test
    trainer = Trainer(
        ctx=mx.context.gpu() if is_gpu & args.use_cuda else mx.context.cpu(),
        batch_size=args.batch_size,
        learning_rate=args.learning_rate,
        epochs=args.max_epochs,
        num_batches_per_epoch=args.number_of_batches_per_epoch,
        clip_gradient=args.clip_gradient,
        weight_decay=args.weight_decay,
        hybridize=True,
    )  # hybridize false for development

    estimator = DeepRenewalEstimator(
        prediction_length=prediction_length,
        context_length=prediction_length * args.context_length_multiplier,
        num_layers=args.num_layers,
        num_cells=args.num_cells,
        cell_type=args.cell_type,
        dropout_rate=args.dropout_rate,
        scaling=True,
        lags_seq=np.arange(1, args.num_lags + 1).tolist(),
        freq=freq,
        use_feat_dynamic_real=args.use_feat_dynamic_real,
        use_feat_static_cat=args.use_feat_static_cat,
        use_feat_static_real=args.use_feat_static_real,
        cardinality=cardinality if args.use_feat_static_cat else None,
        trainer=trainer,
    )
    predictor = estimator.train(train_ds, test_ds)
    os.makedirs(args.model_save_dir, exist_ok=True)
    os.makedirs(Path(args.model_save_dir)/"deep_renewal", exist_ok=True)
    predictor.serialize(Path(args.model_save_dir)/"deep_renewal")
    print("Generating DeepRenewal forecasts.......")
    # predictor = Predictor.deserialize(Path(args.model_save_dir)/"deeprenewal")
    deep_renewal_flat_forecast_it, ts_it = make_evaluation_predictions(
        dataset=test_ds, predictor=predictor, num_samples=100
    )
    tss = list(tqdm(ts_it, total=len(test_ds)))
    deep_renewal_flat_forecasts = list(tqdm(deep_renewal_flat_forecast_it, total=len(test_ds)))
    evaluator = IntermittentEvaluator(quantiles=[0.25,0.5,0.75], median=args.point_forecast=="median", calculate_spec=args.calculate_spec)
    deep_renewal_flat_agg_metrics, deep_renewal_flat_item_metrics = evaluator(
        iter(tss), iter(deep_renewal_flat_forecasts), num_series=len(test_ds)
    )
    print(deep_renewal_flat_agg_metrics)



def main():
    """Console script for deeprenewal."""
    parser = ArgumentParser(description="GluonTS implementation of paper 'Intermittent Demand Forecasting with Deep Renewal Processes'")
    # PROGRAM level args
    parser.add_argument("--use-cuda", type=bool, default=True)
    parser.add_argument("--datasource", type=str, default="retail_dataset", choices=["retail_dataset"])
    parser.add_argument("--regenerate-datasource", help="Whether to discard locally saved dataset and regenerate from source", type=bool, default=False)
    parser.add_argument("--model-save-dir", help="Folder to save models", type=str, default="saved_models")
    parser.add_argument("--point-forecast", help="How to estimate point forecast? Mean or Median", type=str, default="mean", choices=['median', 'mean'])
    parser.add_argument("--calculate-spec", help="Whether to calculate SPEC. It is computationally expensive and therefore False by default", type=bool, default=False)

    # Trainer specific args
    parser.add_argument("--batch_size", type=int, default=32)
    parser.add_argument("--learning-rate", type=float, default=1e-2)
    parser.add_argument("--max-epochs", type=int, default=20)
    parser.add_argument("--number-of-batches-per-epoch", type=int, default=100)
    parser.add_argument("--clip-gradient", type=float, default=5.170127652392614)
    parser.add_argument("--weight-decay", type=float, default=0.01)

    # Model specific args
    parser.add_argument("--context-length-multiplier", help="If context multipler is 2, context available to hte RNN is 2*prediction length", type=int, default=2)
    parser.add_argument("--num-layers", type=int, default=2)
    parser.add_argument("--num-cells", type=int, default=64)
    parser.add_argument("--cell-type", type=str, default="lstm")
    # p% are dropped and set to zero
    parser.add_argument("--dropout-rate", type=float, default=0.3)
    parser.add_argument("--use-feat-dynamic-real", type=bool, default=False)
    parser.add_argument("--use-feat-static-cat", type=bool, default=False)
    parser.add_argument("--use-feat-static-real", type=bool, default=False)
    parser.add_argument("--scaling", help="Whether to scale targets or not", type=bool, default=True)
    parser.add_argument("--num-parallel-samples", type=int, default=100)
    parser.add_argument("--num-lags", help="Number of lags to be included as feature", type=int, default=1)
    parser.add_argument("--forecast-type", help="Defines how the forecast is decoded. For details look at the documentation", type=str, default="hybrid")
    args = parser.parse_args()

    print("Arguments: " + str(args._))
    evaluate_forecast(args)
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/usr/bin/env python

"""Tests for `deeprenewal` package."""


# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# Standard library imports
from typing import Tuple, List

# Third-party imports
import numpy as np
import pandas as pd
import mxnet as mx
import pytest
from random import randint

# First-party imports
import gluonts
from gluonts import time_feature, transform
from gluonts.core import fqname_for
from gluonts.core.serde import dump_code, dump_json, load_code, load_json
from gluonts.dataset.common import ProcessStartField, DataEntry, ListDataset
from gluonts.dataset.field_names import FieldName
from gluonts.evaluation import Evaluator
from gluonts.model.forecast import QuantileForecast, SampleForecast
from gluonts.evaluation.backtest import make_evaluation_predictions 
from gluonts.trainer import Trainer
from deeprenewal import DeepRenewalEstimator, IntermittentEvaluator
from deeprenewal.deeprenewal._transforms import (
    AddInterDemandPeriodFeature,
)

FREQ = "1D"

INTER_DEMAND_TEST_VALUES = {
    "is_train": [True, False],
    "target": [
        (np.zeros(0), np.array([])),
        (np.zeros(13), np.array([1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])),
        (
            np.array([1, 1, 0, 0, 0, 4, 0, 0, 0, 8]),
            np.array([1, 1, 1, 2, 3, 4, 1, 2, 3, 4]),
        ),
    ],
    "start": [
        ProcessStartField.process("2012-01-02", freq="1D"),
    ],
}

ZERO_DEMAND_TEST_VALUES = {
    "is_train": [True, False],
    "target": [
        (np.zeros(0), np.array([])),
        (np.zeros(13), np.array([])),
        (
            np.array([1, 1, 0, 0, 0, 4, 0, 0, 0, 8]),
            np.array([1, 1, 4, 8]),
        ),
    ],
    "start": [
        ProcessStartField.process("2012-01-02", freq="1D"),
    ],
}
#type, hybridize, freq, num_feat_dynamic_real, cardinality
DEEP_RENEWAL_TEST_VALUES = {
    "type":['synthetic','constant'],
    "hybridize": [True, False],
    "freq" :["D","M"],
    "num_feat_dynamic_real": [0,2],
    "cardinality": [[],[2]]
}


@pytest.mark.parametrize("is_train", INTER_DEMAND_TEST_VALUES["is_train"])
@pytest.mark.parametrize("target", INTER_DEMAND_TEST_VALUES["target"])
@pytest.mark.parametrize("start", INTER_DEMAND_TEST_VALUES["start"])
def testAddInterDemandPeriodFeature(start, target, is_train):
    pred_length = 13
    t = AddInterDemandPeriodFeature(
        start_field=FieldName.START,
        target_field=FieldName.TARGET,
        output_field="myout",
        pred_length=pred_length
        # dtype=np.float64,
    )

    assert_serializable(t)

    data = {"start": start, "target": target[0]}
    res = t.map_transform(data, is_train=is_train)
    mat = res["myout"]
    assert np.array_equal(mat, target[1])


@pytest.mark.parametrize("type", DEEP_RENEWAL_TEST_VALUES["type"])
@pytest.mark.parametrize("hybridize", DEEP_RENEWAL_TEST_VALUES["hybridize"])
@pytest.mark.parametrize("freq", DEEP_RENEWAL_TEST_VALUES["freq"])
@pytest.mark.parametrize("num_feat_dynamic_real", DEEP_RENEWAL_TEST_VALUES["num_feat_dynamic_real"])
@pytest.mark.parametrize("cardinality", DEEP_RENEWAL_TEST_VALUES["cardinality"])
def testDeepRenewal(type, hybridize, freq, num_feat_dynamic_real, cardinality):
    prediction_length = 3
    if type == "synthetic":
        train_ds, test_ds = make_dummy_datasets_with_features(
            prediction_length=prediction_length,
            freq=freq,
            num_feat_dynamic_real=num_feat_dynamic_real,
            cardinality=cardinality,
        )
    else:
        train_ds = make_constant_dataset(train_length=15, freq=freq)
        test_ds = train_ds
    trainer = Trainer(
        ctx="cpu", epochs=1, hybridize=hybridize
    )  # hybridize false for development
    estimator = DeepRenewalEstimator(
        prediction_length=prediction_length,
        freq=freq,
        trainer=trainer,
    )
    predictor = estimator.train(training_data=train_ds)
    forecast_it, ts_it = make_evaluation_predictions(
        dataset=test_ds, predictor=predictor, num_samples=100
    )
    evaluator = Evaluator(calculate_owa=False, num_workers=0)

    agg_metrics, item_metrics = evaluator(
        ts_it, forecast_it, num_series=len(test_ds)
    )
    if type == "synthetic":
        accuracy = 1.5
    else:
        accuracy = 1.3
    assert agg_metrics['ND']<= accuracy

def make_dummy_datasets_with_features(
    num_ts: int = 5,
    start: str = "2018-01-01",
    freq: str = "D",
    min_length: int = 5,
    max_length: int = 10,
    prediction_length: int = 3,
    cardinality: List[int] = [],
    num_feat_dynamic_real: int = 0,
) -> Tuple[ListDataset, ListDataset]:

    data_iter_train = []
    data_iter_test = []

    for k in range(num_ts):
        ts_length = randint(min_length, max_length)
        perc_zeros = 0.5
        mask = np.random.rand(ts_length) < perc_zeros
        target = np.array([1.0] * ts_length)
        target[mask] = 0
        data_entry_train = {
            FieldName.START: start,
            FieldName.TARGET: target,
        }
        if len(cardinality) > 0:
            data_entry_train[FieldName.FEAT_STATIC_CAT] = [
                randint(0, c) for c in cardinality
            ]
        data_entry_test = data_entry_train.copy()
        if num_feat_dynamic_real > 0:
            data_entry_train[FieldName.FEAT_DYNAMIC_REAL] = [
                [float(1 + k)] * ts_length for k in range(num_feat_dynamic_real)
            ]
            data_entry_test[FieldName.FEAT_DYNAMIC_REAL] = [
                [float(1 + k)] * (ts_length + prediction_length)
                for k in range(num_feat_dynamic_real)
            ]
        data_iter_train.append(data_entry_train)
        data_iter_test.append(data_entry_test)

    return (
        ListDataset(data_iter=data_iter_train, freq=freq),
        ListDataset(data_iter=data_iter_test, freq=freq),
    )


def make_constant_dataset(train_length,N=5, freq="D"):
    # generates 2 ** N - 1 timeseries with constant increasing values
    n = 2 ** N - 1
    perc_zeros = 0.5
    mask = np.random.rand(train_length) < perc_zeros
    targets = np.ones((n, train_length))
    for i in range(0, n):
        targets[i, :] = targets[i, :] * i
        targets[i, mask] = 0

    ds = gluonts.dataset.common.ListDataset(
        data_iter=[{"start": "2012-01-01", "target": targets[i, :]} for i in range(n)],
        freq=freq,
    )

    return ds


def assert_serializable(x: transform.Transformation):
    t = fqname_for(x.__class__)
    y = load_json(dump_json(x))
    z = load_code(dump_code(x))
    assert dump_json(x) == dump_json(
        y
    ), f"Code serialization for transformer {t} does not work"
    assert dump_code(x) == dump_code(
        z
    ), f"JSON serialization for transformer {t} does not work"


def assert_shape(array: np.array, reference_shape: Tuple[int, int]):
    assert (
        array.shape == reference_shape
    ), f"Shape should be {reference_shape} but found {array.shape}."


def assert_padded_array(
    sampled_array: np.array, reference_array: np.array, padding_array: np.array
):
    num_padded = int(np.sum(padding_array))
    sampled_no_padding = sampled_array[:, num_padded:]

    reference_array = np.roll(reference_array, num_padded, axis=1)
    reference_no_padding = reference_array[:, num_padded:]

    # Convert nans to dummy value for assertion because
    # np.nan == np.nan -> False.
    reference_no_padding[np.isnan(reference_no_padding)] = 9999.0
    sampled_no_padding[np.isnan(sampled_no_padding)] = 9999.0

    reference_no_padding = np.array(reference_no_padding, dtype=np.float32)

    assert (sampled_no_padding == reference_no_padding).all(), (
        f"Sampled and reference arrays do not match. '"
        f"Got {sampled_no_padding} but should be {reference_no_padding}."
    )

def data_iterator(ts):
    """
    :param ts: list of pd.Series or pd.DataFrame
    :return:
    """
    for i in range(len(ts)):
        yield ts[i]


def fcst_iterator(fcst, start_dates, freq):
    """
    :param fcst: list of numpy arrays with the sample paths
    :return:
    """
    for i in range(len(fcst)):
        yield SampleForecast(
            samples=fcst[i], start_date=start_dates[i], freq=freq
        )


def iterator(it):
    """
    Convenience function to toggle whether to consume dataset and forecasts as iterators or iterables.
    :param it:
    :return: it (as iterator)
    """
    return iter(it)


def iterable(it):
    """
    Convenience function to toggle whether to consume dataset and forecasts as iterators or iterables.
    :param it:
    :return: it (as iterable)
    """
    return list(it)


def naive_forecaster(ts, prediction_length, num_samples=100, target_dim=0):
    """
    :param ts: pandas.Series
    :param prediction_length:
    :param num_samples: number of sample paths
    :param target_dim: number of axes of target (0: scalar, 1: array, ...)
    :return: np.array with dimension (num_samples, prediction_length)
    """

    # naive prediction: last observed value
    naive_pred = ts.values[-prediction_length - 1]
    assert len(naive_pred.shape) == target_dim
    return np.tile(
        naive_pred,
        (num_samples, prediction_length) + tuple(1 for _ in range(target_dim)),
    )

def naive_multivariate_forecaster(ts, prediction_length, num_samples=100):
    return naive_forecaster(ts, prediction_length, num_samples, target_dim=1)


def calculate_metrics(
    timeseries,
    evaluator,
    ts_datastructure,
    has_nans=False,
    forecaster=naive_forecaster,
    input_type=iterator,
):
    num_timeseries = timeseries.shape[0]
    num_timestamps = timeseries.shape[1]

    if has_nans:
        timeseries[0, 1] = np.nan
        timeseries[0, 7] = np.nan

    num_samples = 100
    prediction_length = 3
    freq = "1D"

    ts_start_dates = (
        []
    )  # starting date of each time series - can be different in general
    pd_timeseries = []  # list of pandas.DataFrame
    samples = []  # list of forecast samples
    start_dates = []  # start date of the prediction range
    for i in range(num_timeseries):
        ts_start_dates.append(pd.Timestamp(year=2018, month=1, day=1, hour=1))
        index = pd.date_range(
            ts_start_dates[i], periods=num_timestamps, freq=freq
        )

        pd_timeseries.append(ts_datastructure(timeseries[i], index=index))
        samples.append(
            forecaster(pd_timeseries[i], prediction_length, num_samples)
        )
        start_dates.append(
            pd.date_range(
                ts_start_dates[i], periods=num_timestamps, freq=freq
            )[-prediction_length]
        )

    # data iterator
    data_iter = input_type(data_iterator(pd_timeseries))
    fcst_iter = input_type(fcst_iterator(samples, start_dates, freq))

    # evaluate
    agg_df, item_df = evaluator(data_iter, fcst_iter)
    return agg_df, item_df

QUANTILES = [str(q / 10.0) for q in range(1, 10)]

TIMESERIES = [
    np.zeros((5, 10), dtype=np.float64),
    np.ones((5, 10), dtype=np.float64),
    np.ones((5, 10), dtype=np.float64),
    np.arange(0, 50, dtype=np.float64).reshape(5, 10),
]

RES = [
    {
        "MSE": 0.0,
        "abs_error": 0.0,
        "abs_target_sum": 0.0,
        "abs_target_mean": 0.0,
        "seasonal_error": 0.0,
        "MAPE": 0.0,
        "RMSE": 0.0,
        "CFE": 0,
        "PIS": 0,
        "SPEC_0.75": 0,
        "n":  15

    },
    {
        "MSE": 0.0,
        "abs_error": 0.0,
        "abs_target_sum": 15.0,
        "abs_target_mean": 1.0,
        "seasonal_error": 0.0,
        "MAPE": 0.0,
        "RMSE": 0.0,
        "CFE": 0,
        "PIS": 0,
        "SPEC_0.75": 0,
        "n": 15
    },
    {
        "MSE": 0.0,
        "abs_error": 0.0,
        "abs_target_sum": 14.0,
        "abs_target_mean": 1.0,
        "seasonal_error": 0.0,
        "MAPE": 0.0,
        "RMSE": 0.0,
        "CFE": 0,
        "PIS": 0,
        "SPEC_0.75": 0,
        "n": 15,
        "non_zero_n": 14
    },
    {
        "MSE": 4.666666666666,
        "abs_error": 30.0,
        "abs_target_sum": 420.0,
        "abs_target_mean": 28.0,
        "seasonal_error": 6.0,
        "MASE": 1.0,
        "MAPE": 0.10311221153252485,
        "MAAPE": 0.10176599413474399,
        "RMSE": 2.160246899469287,
        "MRAE": 1.0,
        "CFE": 30.,
        "PIS": -50.,
        "NOSp": 1.,
        "SPEC_0.75": 2.5,
        "n": 15,
        "non_zero_n": 15,
        "RelRMSE": 0.5773502691896258,
        "RelMAE": 0.3333333333333333,
        "PBMAE": 0.0
    },
]

HAS_NANS = [False, False, True, False]


INPUT_TYPE = [iterable, iterable, iterable, iterator]


@pytest.mark.parametrize(
    "timeseries, res, has_nans, input_type",
    zip(TIMESERIES, RES, HAS_NANS, INPUT_TYPE),
)
def test_metrics(timeseries, res, has_nans, input_type):
    ts_datastructure = pd.Series
    evaluator = IntermittentEvaluator(quantiles=QUANTILES, num_workers=None, calculate_spec=True)
    agg_metrics, item_metrics = calculate_metrics(
        timeseries,
        evaluator,
        ts_datastructure,
        has_nans=has_nans,
        input_type=input_type,
    )

    for metric, score in agg_metrics.items():
        if metric in res.keys():
            assert abs(score - res[metric]) < 0.001, (
                "Scores for the metric {} do not match: \nexpected: {} "
                "\nobtained: {}".format(metric, res[metric], score)
            )

# for (timeseries, res, has_nans, input_type) in zip(TIMESERIES, RES, HAS_NANS, INPUT_TYPE):
#     test_metrics(timeseries, res, has_nans, input_type)

# testDropNonZeroTarget(
#     start=ProcessStartField.process("2012-01-02", freq="1D"),
#     target=(np.zeros(13), np.array([])),
#     is_train=False,
# )

# testDeepRenewal("constant", True)


from typing import Dict, Iterable, Optional, Tuple, Union

import numpy as np
import pandas as pd
from gluonts.evaluation import Evaluator
from gluonts.model.forecast import Forecast, Quantile


class IntermittentEvaluator(Evaluator):
    """An Evaluator which implements metrics which are more attuned to be used in intermittent
    demand patterns

    Parameters:
        quantiles: list of strings of the form 'p10' or floats in [0, 1] with
            the quantile levels
        seasonality: seasonality to use for seasonal_error, if nothing is passed
            uses the default seasonality
            for the given series frequency as returned by `get_seasonality`
        alpha: Parameter of the MSIS metric from the M4 competition that
            defines the confidence interval.
            For alpha=0.05 (default) the 95% considered is considered in the metric,
            see https://www.m4.unic.ac.cy/wp-content/uploads/2018/03/M4-Competitors-Guide.pdf
            for more detail on MSIS
        calculate_owa: Determines whether the OWA metric should also be calculated,
            which is computationally expensive to evaluate and thus slows
            down the evaluation process considerably.
            By default False.
        calculate_spec: Determines whether the SPEC metric should also be calculated,
            which is computationally expensive to evaluate and thus slows
            down the evaluation process considerably.
            By default False.
        median: Determines whether to use median or mean for point estimation
            By default True
        round_integer: Determines whether to round the forecasts to nearest digit before evaluating forecasts
            By default True..
        num_workers: The number of multiprocessing workers that will be used to process
            the data in parallel.
            Default is multiprocessing.cpu_count().
            Setting it to 0 means no multiprocessing.
        chunk_size: Controls the approximate chunk size each workers handles at a time.
            Default is 32.
    """

    default_quantiles = 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9

    def __init__(
        self,
        quantiles: Iterable[Union[float, str]] = default_quantiles,
        seasonality: Optional[int] = None,
        alpha: float = 0.05,
        calculate_owa: bool = False,
        num_workers: Optional[int] = None,
        chunk_size: Optional[int] = None,
        median: Optional[bool] = True,
        calculate_spec: Optional[bool] = False,
        round_integer: Optional[bool] = False,
    ) -> None:
        self.median = median
        self.calculate_spec = calculate_spec
        self.round_integer = round_integer

        super().__init__(
            quantiles, seasonality, alpha, calculate_owa, num_workers, chunk_size
        )

    @staticmethod
    def naive_fc(
        time_series: Union[pd.Series, pd.DataFrame], forecast: Forecast
    ) -> np.ndarray:
        """

        Parameters:
        ----------
            time_series:time_series
            forecast:forecast

        Returns:
        -------
            np.ndarray
                time series without the forecast dates
        """

        assert forecast.index.intersection(time_series.index).equals(forecast.index), (
            "Index of forecast is outside the index of target\n"
            f"Index of forecast: {forecast.index}\n Index of target: {time_series.index}"
        )
        # Extending the last demand to the prediction length
        date_before_forecast = forecast.index[0] - forecast.index[0].freq
        naive_fc = time_series.loc[date_before_forecast]
        if isinstance(naive_fc, pd.Series):
            naive_fc = naive_fc.values.item()
        return np.atleast_1d(
            np.squeeze(pd.Series(index=forecast.index, data=naive_fc).transpose())
        )

    @staticmethod
    def signed_error(target: Union[pd.Series, np.ndarray], forecast: Union[pd.Series, np.ndarray]):
        """Calculates the signed error (Ground Truth - Forecast)

        Args:
            target (Union[pd.Series, np.ndarray]): Ground Truth
            forecast (Union[pd.Series, np.ndarray]): Forecast

        Returns:
            float: The total signed error
        """
        return np.sum(target - forecast)

    @staticmethod
    def cum_error(target: Union[pd.Series, np.ndarray], forecast: Union[pd.Series, np.ndarray]):
        """Cumulative Error

        Args:
            target (Union[pd.Series, np.ndarray]): Ground Truth
            forecast (Union[pd.Series, np.ndarray]): Forecast

        Returns:
            float: cumulative error
        """
        return np.cumsum(target - forecast)

    @staticmethod
    def nos_p(target: Union[pd.Series, np.ndarray], forecast: Union[pd.Series, np.ndarray]):
        """Calculates the percent of periods which are stock out
            # of periods where CFE > 0 and Ground Truth >0 upon total
            # number of periods 

        Args:
            target (Union[pd.Series, np.ndarray]): Ground Truth
            forecast (Union[pd.Series, np.ndarray]): Forecast

        Returns:
            float: Percentage of periods which is stockout
        """
        c = np.cumsum(target - forecast)
        mask = target > 0
        return np.sum(c[mask] > 0) / np.sum(mask)

    @staticmethod
    def pis(target: Union[pd.Series, np.ndarray], forecast: Union[pd.Series, np.ndarray]):
        """Measures the total number of periods the forecasted item spends in 
            stock or number of stock out periods. It can also be calculated as 
            the cumulative sum of the CFE

        Args:
            target (Union[pd.Series, np.ndarray]): Ground Truth
            forecast (Union[pd.Series, np.ndarray]): Forecast

        Returns:
            float: PIS
        """
        cfe_t = np.cumsum(target - forecast)
        return np.sum(-1 * cfe_t)

    @staticmethod
    def mae(target: Union[pd.Series, np.ndarray], forecast: Union[pd.Series, np.ndarray]):
        """Calculated Mean Absolute Error (mean(abs(ground_truth - forecast)))

        Args:
            target (Union[pd.Series, np.ndarray]): Ground Truth
            forecast (Union[pd.Series, np.ndarray]): Forecast

        Returns:
            float: MAE
        """
        return np.mean(np.abs(target - forecast))

    @staticmethod
    def mean_relative_abs_error(target: Union[pd.Series, np.ndarray], forecast: Union[pd.Series, np.ndarray], naive_fc: Union[pd.Series, np.ndarray]):
        """Mean Absolute Relative Error = Absolute Error/ Abs(Ground Truth - ReferenceFC)

        Args:
            target (Union[pd.Series, np.ndarray]): Ground Truth
            forecast (Union[pd.Series, np.ndarray]): Forecast
            naive_fc (Union[pd.Series, np.ndarray]): reference forecast, most commonly naive forecast

        Returns:
            float: MRAE
        """
        return np.mean(np.abs(target - forecast) / np.abs(target - naive_fc))

    @staticmethod
    def percent_better(target: Union[pd.Series, np.ndarray], forecast: Union[pd.Series, np.ndarray], naive_fc: Union[pd.Series, np.ndarray]):
        """Percent Better finds the percent of instances where the MAE is better than reference MAE

        Args:
            target (Union[pd.Series, np.ndarray]): Ground Truth
            forecast (Union[pd.Series, np.ndarray]): Forecast
            naive_fc (Union[pd.Series, np.ndarray]): Reference forecast, most commonly naive forecast

        Returns:
            float: PBMAE
        """
        mae = np.abs(target - forecast)
        mae_star = np.abs(target - naive_fc)
        pb = mae > mae_star
        return np.sum(pb) / len(pb)
    
    @staticmethod
    #https://github.com/DominikMartin/spec_metric/blob/master/spec_metric/_metric.py
    def spec(y_true, y_pred, a1=0.75, a2=0.25):
        """Stock-keeping-oriented Prediction Error Costs (SPEC)
        Read more in the :ref:`https://arxiv.org/abs/2004.10537`.
        Parameters:
            y_true : array-like of shape (n_samples,)
                Ground truth (correct) target values.
            y_pred : array-like of shape (n_samples,)
                Estimated target values.
            a1 : opportunity costs weighting parameter
                a1 ∈ [0, ∞]. Default value is 0.75.
            a2 : stock-keeping costs weighting parameter
                a2 ∈ [0, ∞]. Default value is 0.25.
        Returns:
        -------
            loss : float
                SPEC output is non-negative floating point. The best value is 0.0.
        Examples:
        --------
            >>> from spec_metric import spec
            >>> y_true = [0, 0, 5, 6, 0, 5, 0, 0, 0, 8, 0, 0, 6, 0]
            >>> y_pred = [0, 0, 5, 6, 0, 5, 0, 0, 8, 0, 0, 0, 6, 0]
            >>> spec(y_true, y_pred)
            0.1428...
            >>> spec(y_true, y_pred, a1=0.1, a2=0.9)
            0.5142...
        """
        assert len(y_true) > 0 and len(y_pred) > 0
        assert len(y_true) == len(y_pred)

        sum_n = 0
        for t in range(1, len(y_true) + 1):
            sum_t = 0
            for i in range(1, t + 1):
                delta1 = np.sum([y_k for y_k in y_true[:i]]) - np.sum([f_j for f_j in y_pred[:t]])
                delta2 = np.sum([f_k for f_k in y_pred[:i]]) - np.sum([y_j for y_j in y_true[:t]])

                sum_t = sum_t + np.max([0, a1 * np.min([y_true[i - 1], delta1]), a2 * np.min([y_pred[i - 1], delta2])]) * (
                            t - i + 1)
            sum_n = sum_n + sum_t
        return sum_n / len(y_true)
    
    
    @staticmethod
    def maape(target: Union[pd.Series, np.ndarray], forecast: Union[pd.Series, np.ndarray]):
        """Mean Arctangent Absolute Percent Error is a variant of MAPE which is not affected by division by zero.
            https://www.sciencedirect.com/science/article/pii/S0169207016000121

        Args:
            target (Union[pd.Series, np.ndarray]): Ground Truth
            forecast (Union[pd.Series, np.ndarray]): Forecast

        Returns:
            float: MAAPE
        """
        ape = np.zeros_like(target, dtype="float")
        mask = np.logical_not((target == 0) & (forecast == 0))
        ape = np.divide(np.abs(target - forecast), target, out=ape, where=mask)
        return np.mean(np.arctan(ape))

    def get_metrics_per_ts(
        self, time_series: Union[pd.Series, pd.DataFrame], forecast: Forecast
    ) -> Dict[str, Union[float, str, None]]:
        pred_target = np.array(self.extract_pred_target(time_series, forecast))
        pred_target = np.ma.masked_invalid(pred_target)

        # required for seasonal_error and owa calculation
        past_data = np.array(self.extract_past_data(time_series, forecast))
        past_data = np.ma.masked_invalid(past_data)
        if self.median:
            fcst = forecast.quantile(0.5)
        else:
            fcst = forecast.mean
        if self.round_integer:
            fcst = np.round(fcst)
        naive_fc = self.naive_fc(time_series, forecast)
        seasonal_error = np.abs(naive_fc - pred_target)
        seasonal_sq_error = np.square(naive_fc - pred_target)
        nonzero_mask = pred_target > 0
        metrics = {
            "item_id": forecast.item_id,
            "non_zero_n": np.sum(nonzero_mask),
            "n": len(pred_target),
            "abs_target_sum": self.abs_target_sum(pred_target),
            "abs_target_mean": self.abs_target_mean(pred_target),
            "seasonal_error": seasonal_error.sum(),
            "seasonal_sq_error": seasonal_sq_error.sum(),
            "abs_error": self.abs_error(pred_target, fcst),
            "cum_error": self.cum_error(pred_target, fcst),
            "CFE": self.signed_error(pred_target, fcst),
            "CFE_min": np.min(self.cum_error(pred_target, fcst)),
            "CFE_max": np.max(self.cum_error(pred_target, fcst)),
            "NOSp": self.nos_p(pred_target, fcst),
            "PIS": self.pis(pred_target, fcst),
            "MSE": self.mse(pred_target, fcst),
            "RMSE": np.sqrt(self.mse(pred_target, fcst)),
            "MAE": self.mae(pred_target, fcst),
            "MRAE": self.mean_relative_abs_error(pred_target, fcst, naive_fc),
            "MASE": self.mase(pred_target, fcst, seasonal_error.mean()),
            "MAPE": self.mape(pred_target, fcst),
            "RelMAE": self.mae(pred_target, fcst) / self.mae(pred_target, naive_fc),
            "RelRMSE": np.sqrt(self.mse(pred_target, fcst))
            / np.sqrt(self.mse(pred_target, naive_fc)),
            "PBMAE": self.percent_better(pred_target, fcst, naive_fc),
            "MAAPE": self.maape(pred_target, fcst),
            "sMAPE": self.smape(pred_target, fcst),
        }
        if self.calculate_spec:
            metrics["SPEC_0.75"] = self.spec(pred_target, fcst, a1=0.75, a2=0.25)
            metrics["SPEC_0.5"] = self.spec(pred_target, fcst, a1=0.5, a2=0.5)
            metrics["SPEC_0.25"] = self.spec(pred_target, fcst, a1=0.25, a2=0.75)
        for quantile in self.quantiles:
            forecast_quantile = forecast.quantile(quantile.value)

            metrics[quantile.loss_name] = self.quantile_loss(
                pred_target, forecast_quantile, quantile.value
            )
            metrics[quantile.coverage_name] = self.coverage(
                pred_target, forecast_quantile
            )

        return metrics

    def get_aggregate_metrics(
        self, metric_per_ts: pd.DataFrame
    ) -> Tuple[Dict[str, float], pd.DataFrame]:
        agg_funs = {
            "MSE": "mean",
            "MAE": "mean",
            "MRAE": "mean",
            "MASE": "mean",
            "MAPE": "mean",
            "MAAPE": "mean",
            "sMAPE": "mean",
            "abs_target_sum": "sum",
            "abs_target_mean": "mean",
            "abs_error": "sum",
            "seasonal_error": "mean",
            "n": "sum",
            "non_zero_n": "sum",
            "CFE": "sum",
            "PIS": "sum",
        }
        for quantile in self.quantiles:
            agg_funs[quantile.loss_name] = "sum"
            agg_funs[quantile.coverage_name] = "mean"

        assert (
            set(metric_per_ts.columns) >= agg_funs.keys()
        ), "The some of the requested item metrics are missing."

        totals = {key: metric_per_ts[key].agg(agg) for key, agg in agg_funs.items()}

        # derived metrics based on previous aggregate metrics
        totals["NOSp"] = (
            metric_per_ts["NOSp"] * metric_per_ts["non_zero_n"]
        ).sum() / metric_per_ts["non_zero_n"].sum()
        totals["PBMAE"] = (
            metric_per_ts["PBMAE"] * metric_per_ts["n"]
        ).sum() / metric_per_ts["n"].sum()
        if self.calculate_spec:
            totals["SPEC_0.75"] = (
                metric_per_ts["SPEC_0.75"] * metric_per_ts["n"]
            ).sum() / metric_per_ts["n"].sum()
            totals["SPEC_0.5"] = (
                metric_per_ts["SPEC_0.5"] * metric_per_ts["n"]
            ).sum() / metric_per_ts["n"].sum()
            totals["SPEC_0.25"] = (
                metric_per_ts["SPEC_0.25"] * metric_per_ts["n"]
            ).sum() / metric_per_ts["n"].sum()
        totals["RMSE"] = np.sqrt(totals["MSE"])
        cum_errs = np.array([row for row in metric_per_ts["cum_error"]])
        cum_errs = np.sum(cum_errs, axis=0)
        totals["CFE_min"] = cum_errs.min()
        totals["CFE_max"] = cum_errs.max()
        totals["RelMAE"] = totals["MAE"] / metric_per_ts["seasonal_error"].mean()
        totals["RelRMSE"] = totals["RMSE"] / np.sqrt(
            metric_per_ts["seasonal_sq_error"].mean()
        )

        flag = totals["abs_target_mean"] == 0
        totals["NRMSE"] = np.divide(
            totals["RMSE"] * (1 - flag), totals["abs_target_mean"] + flag
        )

        flag = totals["abs_target_sum"] == 0
        totals["ND"] = np.divide(
            totals["abs_error"] * (1 - flag), totals["abs_target_sum"] + flag
        )

        all_qLoss_names = [quantile.weighted_loss_name for quantile in self.quantiles]
        for quantile in self.quantiles:
            totals[quantile.weighted_loss_name] = np.divide(
                totals[quantile.loss_name], totals["abs_target_sum"]
            )

        totals["mean_wQuantileLoss"] = np.array(
            [totals[ql] for ql in all_qLoss_names]
        ).mean()

        totals["MAE_Coverage"] = np.mean(
            [
                np.abs(totals[q.coverage_name] - np.array([q.value]))
                for q in self.quantiles
            ]
        )
        return totals, metric_per_ts



# Heavily inspired by GluonTS Implementation
from functools import lru_cache
from typing import Iterator, List, Optional

import numpy as np
import pandas as pd
from gluonts.core.component import validated
from gluonts.core.exception import GluonTSDateBoundsError
from gluonts.dataset.common import DataEntry
from gluonts.dataset.field_names import FieldName
from gluonts.transform._base import FlatMapTransformation
from gluonts.transform.sampler import ContinuousTimePointSampler, InstanceSampler


def shift_timestamp(ts: pd.Timestamp, offset: int) -> pd.Timestamp:
    """
    Computes a shifted timestamp.

    Basic wrapping around pandas ``ts + offset`` with caching and exception
    handling.
    """
    return _shift_timestamp_helper(ts, ts.freq, offset)


@lru_cache(maxsize=10000)
def _shift_timestamp_helper(ts: pd.Timestamp, freq: str, offset: int) -> pd.Timestamp:
    """
    We are using this helper function which explicitly uses the frequency as a
    parameter, because the frequency is not included in the hash of a time
    stamp.

    I.e.
      pd.Timestamp(x, freq='1D')  and pd.Timestamp(x, freq='1min')

    hash to the same value.
    """
    try:
        # this line looks innocent, but can create a date which is out of
        # bounds values over year 9999 raise a ValueError
        # values over 2262-04-11 raise a pandas OutOfBoundsDatetime
        return ts + offset * ts.freq
    except (ValueError, pd._libs.OutOfBoundsDatetime) as ex:
        raise GluonTSDateBoundsError(ex)


class RenewalInstanceSplitter(FlatMapTransformation):
    """
    Selects training instances, after removing the zero demand cases from the
    training sequence, by slicing the target and other time series
    like arrays at random points in training mode or at the last time point in
    prediction mode. Assumption is that all time like arrays start at the same
    time point.

    The target and each time_series_field is removed and instead two
    corresponding fields with prefix `past_` and `future_` are included. E.g.

    If the target array is one-dimensional, the resulting instance has shape
    (len_target). In the multi-dimensional case, the instance has shape (dim,
    len_target).

    target -> past_target and future_target

    The transformation also adds a field 'past_is_pad' that indicates whether
    values where padded or not.

    Convention: time axis is always the last axis.

    Parameters
    ----------

    target_field
        field containing the target
    is_pad_field
        output field indicating whether padding happened
    start_field
        field containing the start date of the time series
    forecast_start_field
        output field that will contain the time point where the forecast starts
    train_sampler
        instance sampler that provides sampling indices given a time-series
    past_length
        length of the target seen before making prediction
    future_length
        length of the target that must be predicted
    lead_time
        gap between the past and future windows (default: 0)
    output_NTC
        whether to have time series output in (time, dimension) or in
        (dimension, time) layout (default: True)
    time_series_fields
        fields that contains time-series, they are split in the same interval
        as the target (default: None)
    pick_incomplete
        whether training examples can be sampled with only a part of
        past_length time-units
        present for the time series. This is useful to train models for
        cold-start. In such case, is_pad_out contains an indicator whether
        data is padded or not. (default: True)
    dummy_value
        Value to use for padding. (default: 0.0)
    """

    @validated()
    def __init__(
        self,
        target_field: str,
        is_pad_field: str,
        start_field: str,
        forecast_start_field: str,
        train_sampler: InstanceSampler,
        past_length: int,
        future_length: int,
        lead_time: int = 0,
        output_NTC: bool = True,
        time_series_fields: Optional[List[str]] = None,
        pick_incomplete: bool = True,
        dummy_value: float = 0.0,
    ) -> None:

        assert future_length > 0

        self.train_sampler = train_sampler
        self.past_length = past_length
        self.future_length = future_length
        self.lead_time = lead_time
        self.output_NTC = output_NTC
        self.ts_fields = time_series_fields if time_series_fields is not None else []
        self.target_field = target_field
        self.is_pad_field = is_pad_field
        self.start_field = start_field
        self.forecast_start_field = forecast_start_field
        self.pick_incomplete = pick_incomplete
        self.dummy_value = dummy_value

    def _past(self, col_name):
        return f"past_{col_name}"

    def _future(self, col_name):
        return f"future_{col_name}"

    @staticmethod
    def remove_zero_demand(
        data: DataEntry, target_field, input_fields, pred_length
    ) -> DataEntry:
        target = (
            data[target_field].reshape(1, -1)
            if data[target_field].ndim == 1
            else data[target_field]
        )
        mask = target[0, :] > 0
        data[target_field] = (
            data[target_field][mask]
            if data[target_field].ndim == 1
            else data[target_field][:, mask]
        )
        # Adding Trues for the prediction length. Useful while prediction
        mask = np.append(mask, [True] * pred_length)
        for field in input_fields:
            _mask = mask[: data[field].shape[-1]]
            if data[field].ndim == 1:
                data[field] = data[field][_mask]
            elif data[field].ndim == 2:
                data[field] = data[field][:, _mask]
            else:
                raise NotImplementedError("ndim for {} should be atmost 2")

        return data

    def flatmap_transform(self, data: DataEntry, is_train: bool) -> Iterator[DataEntry]:
        pl = self.future_length
        lt = self.lead_time
        slice_cols = self.ts_fields + [self.target_field]
        true_target = data[self.target_field]
        true_len_target = true_target.shape[-1]
        data = self.remove_zero_demand(
            data,
            target_field=self.target_field,
            input_fields=[FieldName.FEAT_TIME, FieldName.OBSERVED_VALUES],
            pred_length=self.future_length,
        )

        target = data[self.target_field]

        len_target = target.shape[-1]

        minimum_length = (
            self.future_length
            if self.pick_incomplete
            else self.past_length + self.future_length
        ) + self.lead_time

        if is_train:
            sampling_bounds = (
                (
                    0,
                    len_target - self.future_length - self.lead_time,
                )  # TODO: create parameter lower sampling bound for NBEATS
                if self.pick_incomplete
                else (
                    self.past_length,
                    len_target - self.future_length - self.lead_time,
                )
            )

            # We currently cannot handle time series that are
            # too short during training, so we just skip these.
            # If we want to include them we would need to pad and to
            # mask the loss.
            sampled_indices = (
                np.array([], dtype=int)
                if len_target < minimum_length
                else self.train_sampler(target, *sampling_bounds)
            )
        else:
            assert self.pick_incomplete or len_target >= self.past_length
            sampled_indices = np.array([len_target], dtype=int)
        for i in sampled_indices:
            pad_length = max(self.past_length - i, 0)
            if not self.pick_incomplete:
                assert pad_length == 0, f"pad_length should be zero, got {pad_length}"
            d = data.copy()
            for ts_field in slice_cols:
                if i > self.past_length:
                    # truncate to past_length
                    past_piece = d[ts_field][..., i - self.past_length : i]
                elif i < self.past_length:
                    pad_block = (
                        np.ones(
                            d[ts_field].shape[:-1] + (pad_length,),
                            dtype=d[ts_field].dtype,
                        )
                        * self.dummy_value
                    )
                    past_piece = np.concatenate(
                        [pad_block, d[ts_field][..., :i]], axis=-1
                    )
                else:
                    past_piece = d[ts_field][..., :i]
                d[self._past(ts_field)] = past_piece
                d[self._future(ts_field)] = d[ts_field][..., i + lt : i + lt + pl]
                del d[ts_field]
            pad_indicator = np.zeros(self.past_length)
            if pad_length > 0:
                pad_indicator[:pad_length] = 1

            if self.output_NTC:
                for ts_field in slice_cols:
                    d[self._past(ts_field)] = d[self._past(ts_field)].transpose()
                    d[self._future(ts_field)] = d[self._future(ts_field)].transpose()

            d[self._past(self.is_pad_field)] = pad_indicator
            if is_train:
                pass
                # d[self.forecast_start_field] = shift_timestamp(d[self.start_field], i + lt)
            else:
                d[self.forecast_start_field] = shift_timestamp(
                    d[self.start_field], true_len_target + lt
                )
            yield d



# Heavily inspired by https://github.com/HamidM6/croston/blob/master/croston/croston.py
# Added the option to do separate parameters for demand and intermittent processes
"""
croston model for intermittent time series
"""

import numpy as np
import pandas as pd
from scipy.optimize import minimize


def fit_croston(
    input_endog, forecast_length, croston_variant="original", number_parameters=2
):
    """
    :param input_endog: numpy array of intermittent demand time series
    :param forecast_length: forecast horizon
    :param croston_variant: croston model type
    :return: dictionary of model parameters, in-sample forecast, and out-of-sample forecast
    """

    input_series = np.asarray(input_endog)
    epsilon = 1e-7
    input_length = len(input_series)
    non_zero_demand = np.where(input_series != 0)[0]

    if list(non_zero_demand) != [0]:

        try:
            alpha_opt = _croston_opt(
                input_series=input_series,
                input_series_length=input_length,
                croston_variant=croston_variant,
                epsilon=epsilon,
                alpha=None,
                number_parameters=number_parameters,
            )

            croston_training_result = _croston(
                input_series=input_series,
                input_series_length=input_length,
                croston_variant=croston_variant,
                alpha=alpha_opt,
                horizon=forecast_length,
                epsilon=epsilon,
            )
            croston_model = croston_training_result["model"]
            croston_fittedvalues = croston_training_result["in_sample_forecast"]

            croston_forecast = croston_training_result["out_of_sample_forecast"]

        except Exception as e:

            croston_model = None
            croston_fittedvalues = None
            croston_forecast = None
            print(str(e))

    else:

        croston_model = None
        croston_fittedvalues = None
        croston_forecast = None

    return {
        "croston_model": croston_model,
        "croston_fittedvalues": croston_fittedvalues,
        "croston_forecast": croston_forecast,
    }


def _croston(
    input_series, input_series_length, croston_variant, alpha, horizon, epsilon
):

    # Croston decomposition
    non_zero_demand = np.where(input_series != 0)[0]  # find location of non-zero demand

    k = len(non_zero_demand)
    z = input_series[non_zero_demand]  # demand

    x = np.concatenate([[non_zero_demand[0]], np.diff(non_zero_demand)])  # intervals

    # initialize

    init = [z[0], np.mean(x)]

    zfit = np.array([None] * k)
    xfit = np.array([None] * k)

    # assign initial values and prameters

    zfit[0] = init[0]
    xfit[0] = init[1]

    if len(alpha) == 1:
        a_demand = alpha[0]
        a_interval = alpha[0]

    else:
        a_demand = alpha[0]
        a_interval = alpha[1]

    # compute croston variant correction factors
    #   sba: syntetos-boylan approximation
    #   sbj: shale-boylan-johnston
    #   tsb: teunter-syntetos-babai

    if croston_variant == "sba":
        correction_factor = 1 - (a_interval / 2)

    elif croston_variant == "sbj":
        correction_factor = 1 - a_interval / (2 - a_interval + epsilon)

    else:
        correction_factor = 1

    # fit model

    for i in range(1, k):
        zfit[i] = zfit[i - 1] + a_demand * (z[i] - zfit[i - 1])  # demand
        xfit[i] = xfit[i - 1] + a_interval * (x[i] - xfit[i - 1])  # interval

    cc = correction_factor * zfit / (xfit + epsilon)

    croston_model = {
        "a_demand": a_demand,
        "a_interval": a_interval,
        "demand_series": pd.Series(zfit),
        "interval_series": pd.Series(xfit),
        "demand_process": pd.Series(cc),
        "correction_factor": correction_factor,
    }

    # calculate in-sample demand rate

    frc_in = np.zeros(input_series_length)
    tv = np.concatenate(
        [non_zero_demand, [input_series_length]]
    )  # Time vector used to create frc_in forecasts

    for i in range(k):
        frc_in[tv[i] : min(tv[i + 1], input_series_length)] = cc[i]

    # forecast out_of_sample demand rate

    if horizon > 0:
        frc_out = np.array([cc[k - 1]] * horizon)

    else:
        frc_out = None

    return_dictionary = {
        "model": croston_model,
        "in_sample_forecast": frc_in,
        "out_of_sample_forecast": frc_out,
    }

    return return_dictionary


def _croston_opt(
    input_series,
    input_series_length,
    croston_variant,
    epsilon,
    alpha=None,
    number_parameters=1,
):

    p0 = np.array([0.1] * number_parameters)

    wopt = minimize(
        fun=_croston_cost,
        x0=p0,
        method="Nelder-Mead",
        args=(input_series, input_series_length, croston_variant, epsilon),
    )

    constrained_wopt = np.minimum([1], np.maximum([0], wopt.x))

    return constrained_wopt


def _croston_cost(p0, input_series, input_series_length, croston_variant, epsilon):

    # cost function for croston and variants

    frc_in = _croston(
        input_series=input_series,
        input_series_length=input_series_length,
        croston_variant=croston_variant,
        alpha=p0,
        horizon=0,
        epsilon=epsilon,
    )["in_sample_forecast"]

    E = input_series - frc_in
    E = E[E != np.array(None)]
    E = np.mean(E ** 2)

    return E



from typing import List

import numpy as np
import pandas as pd
from gluonts.core.component import DType, validated
from gluonts.dataset.common import DataEntry
from gluonts.transform import (
    AddTimeFeatures,
    MapTransformation,
    SimpleTransformation,
    shift_timestamp,
    target_transformation_length,
)


# TODO See if inheritance is possible from add time features
class AddInterDemandPeriodFeature(MapTransformation):
    """
    Calculates the Inter Demand Period feature

    Parameters
    ----------
    start_field
        Field with the start time stamp of the time series
    target_field
        Field with the array containing the time series values
    output_field
        Field name for result.
    pred_length
        Prediction length
    """

    @validated()
    def __init__(
        self,
        start_field: str,
        target_field: str,
        output_field: str,
        pred_length: int,
    ) -> None:
        self.pred_length = pred_length
        self.start_field = start_field
        self.target_field = target_field
        self.output_field = output_field
        self._min_time_point: pd.Timestamp = None
        self._max_time_point: pd.Timestamp = None
        self._full_range_date_features: np.ndarray = None
        self._date_index: pd.DatetimeIndex = None

    def _update_cache(self, start: pd.Timestamp, length: int) -> None:
        end = shift_timestamp(start, length)
        if self._min_time_point is not None:
            if self._min_time_point <= start and end <= self._max_time_point:
                return
        if self._min_time_point is None:
            self._min_time_point = start
            self._max_time_point = end
        self._min_time_point = min(shift_timestamp(start, -50), self._min_time_point)
        self._max_time_point = max(shift_timestamp(end, 50), self._max_time_point)
        self.full_date_range = pd.date_range(
            self._min_time_point, self._max_time_point, freq=start.freq
        )
        # self._full_range_date_features = (
        #     np.vstack(
        #         [feat(self.full_date_range) for feat in self.date_features]
        #     )
        #     if self.date_features
        #     else None
        # )
        self._date_index = pd.Series(
            index=self.full_date_range,
            data=np.arange(len(self.full_date_range)),
        )

    def map_transform(self, data: DataEntry, is_train: bool) -> DataEntry:
        start = data[self.start_field]
        length = target_transformation_length(
            data[self.target_field], self.pred_length, is_train=is_train
        )
        self._update_cache(start, length)
        i0 = self._date_index[start]
        date_idx = self._date_index.iloc[i0 : i0 + length].index
        # When is_train is false, date_idx has len of target_len + prediction_len
        # which is useful in time feature generation, but we only need target length
        date_idx = date_idx[: len(data[self.target_field])]
        feature = pd.Series(np.ones(len(date_idx)) * np.nan, index=date_idx)
        mask = data[self.target_field] > 0
        feature.loc[mask] = feature.loc[mask].index
        # filling in nan in first row with the corresponding date
        # Assumption: If the frame starts with a zero demand, earliest date in frame is taken as a start
        if len(feature) > 0:
            if pd.isnull(feature[0]):
                feature[0] = feature.index[0]
        feature = feature.ffill().to_frame()
        # feature["diff"] = (feature.index - feature.iloc[:, 0]) / pd.Timedelta(
        #     1, feature.index.freqstr
        # )
        feature["diff"] = feature.index.to_period(feature.index.freqstr).astype(
            int
        ) - pd.DatetimeIndex(feature.iloc[:, 0]).to_period(
            feature.index.freqstr
        ).astype(
            int
        )
        feature["diff"] = feature["diff"].shift(1).round() + 1
        feature["diff"] = feature["diff"].fillna(method="bfill")
        feature = feature["diff"].values
        if self.output_field in data.keys():
            data[self.output_field] = np.vstack([data[self.output_field], feature])
        else:
            data[self.output_field] = feature
        return data



# Heavily inspired from GluonTS datasets
import json
import logging
import os
from collections import OrderedDict
from functools import partial
from pathlib import Path
from typing import List, Optional

import numpy as np
import pandas as pd
from gluonts.dataset.common import TrainDatasets, load_datasets
from gluonts.dataset.field_names import FieldName
from gluonts.dataset.repository._util import metadata, save_to_file, to_dict
from gluonts.gluonts_tqdm import tqdm
from gluonts.support.util import get_download_path


def _preprocess_retail_data(df, combination):
    df["Country"] = df["Country"].astype("category").cat.codes
    df["StockCode"] = df["StockCode"].astype("category").cat.codes
    df = df.groupby(["StockCode", "Country", "InvoiceDate"]).agg(
        {"Quantity": "sum", "UnitPrice": "mean"}
    )
    counts = df.reset_index().groupby(combination)["StockCode"].count()
    combinations_selected = counts[counts > 10].index
    df = df.reset_index().set_index(combination)
    df = df[df.index.isin(combinations_selected)]
    max_date = df.InvoiceDate.max().replace(hour=0, minute=0, second=0)

    def resample_ds(df):
        df.InvoiceDate = pd.to_datetime(df.InvoiceDate, yearfirst=True)
        df.rename(columns={"InvoiceDate": "date"}, inplace=True)
        new_idx = pd.date_range(
            df.date.min().replace(hour=0, minute=0, second=0),
            max_date,
            freq="1D",
            name="InvoiceDate",
        )
        df.set_index("date", inplace=True)
        df = (
            df.resample("1D")
            .agg({"Quantity": "sum", "UnitPrice": "mean"})
            .reindex(new_idx)
        )
        df["Quantity"] = df["Quantity"].fillna(0)
        df["UnitPrice"] = df["UnitPrice"].ffill().bfill()
        return df

    df = df.reset_index().groupby(combination).apply(resample_ds)
    df.Quantity = df.Quantity.clip(lower=0)
    return df


def generate_retail_dataset(dataset_path: Path, split: str = "2011-11-24"):
    retail_dataset_url = "https://archive.ics.uci.edu/ml/machine-learning-databases/00352/Online%20Retail.xlsx"
    df = pd.read_excel(retail_dataset_url)
    combination = ["StockCode", "Country"]
    df = _preprocess_retail_data(df, combination)
    # df.to_pickle("temp.pkl")
    # df = pd.read_pickle("temp.pkl")
    idx = pd.IndexSlice[:, :, :split]
    train_df = df.loc[idx, :].reset_index()
    idx = pd.IndexSlice[:, :, split:]
    test_df = df.loc[idx, :].reset_index()
    full_df = df.reset_index()
    single_prediction_length = len(test_df["InvoiceDate"].unique())
    feat_static_cat = combination
    feat_dynamic_real = []
    target = "Quantity"
    date_col = "InvoiceDate"

    os.makedirs(dataset_path, exist_ok=True)

    uniq_combs = train_df[combination].drop_duplicates().apply(tuple, axis=1)
    dynamic_real_train_l = []
    dynamic_real_test_l = []
    stat_cat_l = []
    start_l = []
    train_target_l = []
    test_target_l = []
    for stock_code, country in tqdm(uniq_combs):
        df = train_df[
            (train_df.StockCode == stock_code) & (train_df.Country == country)
        ]
        _df = full_df[(full_df.StockCode == stock_code) & (full_df.Country == country)]
        train_ts = _df[target].values.ravel()
        if (train_ts > 0).sum() > (single_prediction_length + 13):
            # test_feat_dyn_array = _df.loc[:, feat_dynamic_real].values.T
            # train_feat_dyn_array = test_feat_dyn_array[:, :-single_prediction_length]

            test_ts = train_ts.copy()
            train_ts = train_ts[:-single_prediction_length]

            # dynamic_real_train_l.append(train_feat_dyn_array)
            # dynamic_real_test_l.append(test_feat_dyn_array)
            start_l.append(df[date_col].min())
            train_target_l.append(train_ts)
            test_target_l.append(test_ts)
            stat_cat_l.append(
                np.squeeze(df.loc[:, feat_static_cat].drop_duplicates().values)
            )
    stat_cat_cardinalities = [len(full_df[col].unique()) for col in feat_static_cat]

    with open(dataset_path / "metadata.json", "w") as f:
        f.write(
            json.dumps(
                metadata(
                    cardinality=stat_cat_cardinalities,
                    freq="1D",
                    prediction_length=single_prediction_length,
                )
            )
        )

    train_file = dataset_path / "train" / "data.json"
    test_file = dataset_path / "test" / "data.json"
    train_ds = [
        {
            FieldName.ITEM_ID: "|".join(map(str, uniq_comb)),
            FieldName.TARGET: target.tolist(),
            FieldName.START: str(start),
            FieldName.FEAT_STATIC_CAT: fsc.tolist(),
            # FieldName.FEAT_DYNAMIC_REAL: fdr.tolist(),
        }
        for uniq_comb, target, start, fdr, fsc in zip(
            uniq_combs,
            train_target_l,
            start_l,
            # dynamic_real_train_l,
            stat_cat_l,
        )
    ]
    save_to_file(train_file, train_ds)
    test_ds = [
        {
            FieldName.ITEM_ID: "|".join(map(str, uniq_comb)),
            FieldName.TARGET: target.tolist(),
            FieldName.START: str(start),
            FieldName.FEAT_STATIC_CAT: fsc.tolist(),
            # FieldName.FEAT_DYNAMIC_REAL: fdr.tolist(),
        }
        for uniq_comb, target, start, fdr, fsc in zip(
            uniq_combs,
            test_target_l,
            start_l,
            # dynamic_real_test_l,
            stat_cat_l,
        )
    ]
    save_to_file(test_file, test_ds)


default_dataset_path = get_download_path() / "datasets"

dataset_recipes = OrderedDict(
    {
        # each recipe generates a dataset given a path
        "retail_dataset": partial(generate_retail_dataset, split="2011-11-01")
    }
)


def materialize_dataset(
    dataset_name: str,
    path: Path = default_dataset_path,
    regenerate: bool = False,
) -> Path:
    """
    Ensures that the dataset is materialized under the `path / dataset_name`
    path.
    Parameters
    ----------
    dataset_name
        name of the dataset, for instance "m4_hourly"
    regenerate
        whether to regenerate the dataset even if a local file is present.
        If this flag is False and the file is present, the dataset will not
        be downloaded again.
    path
        where the dataset should be saved
    Returns
    -------
        the path where the dataset is materialized
    """

    path.mkdir(parents=True, exist_ok=True)
    dataset_path = path / dataset_name

    dataset_recipe = dataset_recipes[dataset_name]

    if not dataset_path.exists() or regenerate:
        logging.info(f"downloading and processing {dataset_name}")
        dataset_recipe(dataset_path=dataset_path)
    else:
        logging.info(f"using dataset already processed in path {dataset_path}.")

    return dataset_path


def get_dataset(
    dataset_name: str,
    path: Optional[Path] = None,
    regenerate: bool = False,
) -> TrainDatasets:
    """
    Get the repository dataset.
    Currently only [Retail Dataset](https://archive.ics.uci.edu/ml/datasets/online+retail) is available
    Parameters:
        dataset_name:
            name of the dataset, for instance "retail"
        regenerate:
            whether to regenerate the dataset even if a local file is present.
            If this flag is False and the file is present, the dataset will not
            be downloaded again.
        path:
            where the dataset should be saved
    Returns:
        dataset obtained by either downloading or reloading from local file.
    """
    if path is None:
        path = default_dataset_path
    dataset_path = materialize_dataset(dataset_name, path, regenerate)

    return load_datasets(
        metadata=dataset_path,
        train=dataset_path / "train",
        test=dataset_path / "test",
    )



"""Top-level package for DeepRenewal."""

__author__ = """Manu Joseph"""
__email__ = 'manujosephv@gmail.com'
__version__ = '0.3.1'

# Relative imports
from .deeprenewal._estimator import DeepRenewalEstimator
from .croston._estimator import CrostonForecastEstimator
from .croston._predictor import CrostonForecastPredictor
from ._datasets import get_dataset
from ._evaluator import IntermittentEvaluator

__all__ = ["get_dataset","DeepRenewalEstimator","CrostonForecastEstimator","CrostonForecastPredictor", "IntermittentEvaluator"]

# fix Sphinx issues, see https://bit.ly/2K2eptM
for item in __all__:
    if hasattr(item, "__module__"):
        setattr(item, "__module__", __name__)


# Heavily inspired by DeepAR implemenation in GluonTS
# Standard library imports
from typing import List, Optional

import numpy as np

# First-party imports
from gluonts.core.component import DType, validated
from gluonts.dataset.field_names import FieldName
from gluonts.distribution import DistributionOutput, NegativeBinomialOutput
from gluonts.model.estimator import GluonEstimator
from gluonts.model.predictor import Predictor, RepresentableBlockPredictor
from gluonts.support.util import copy_parameters
from gluonts.time_feature import (
    TimeFeature,
    get_lags_for_frequency,
    time_features_from_frequency_str,
)
from gluonts.trainer import Trainer
from gluonts.transform import (
    AddAgeFeature,
    AddObservedValuesIndicator,
    AddTimeFeatures,
    AsNumpyArray,
    Chain,
    ExpectedNumInstanceSampler,
    InstanceSplitter,
    RemoveFields,
    SetField,
    Transformation,
    VstackFeatures,
)

# Third-party imports
from mxnet.gluon import HybridBlock

from ._forecast_generator import IntermittentSampleForecastGenerator

# Relative imports
from ._network import DeepRenewalPredictionNetwork, DeepRenewalTrainingNetwork
from ._sampler import RenewalInstanceSplitter
from ._transforms import AddInterDemandPeriodFeature

"""
Construct a DeepRenewal estimator.

This implements an RNN-based model, close to the one described in
https://arxiv.org/abs/1911.10416.

Parameters:

    freq (str): Frequency of the data to train on and predict
    prediction_length (int): Length of the prediction horizon
    trainer (Optional[Trainer]): Trainer object to be used (default: Trainer())
    context_length (Optional[int]): Number of steps to unroll the RNN for before computing predictions (default: None, in which case context_length = prediction_length)
    num_layers: Number of RNN layers (default: 2)
    num_cells: Number of RNN cells for each layer (default: 40)
    cell_type: Type of recurrent cells to use (available: 'lstm' or 'gru';
        default: 'lstm')
    dropout_rate: Dropout regularization parameter (default: 0.1)
    use_feat_dynamic_real: Whether to use the ``feat_dynamic_real`` field from the data
        (default: False)
    use_feat_static_cat: Whether to use the ``feat_static_cat`` field from the data
        (default: False)
    use_feat_static_real: Whether to use the ``feat_static_real`` field from the data
        (default: False)
    cardinality: Number of values of each categorical feature.
        This must be set if ``use_feat_static_cat == True`` (default: None)
    embedding_dimension: Dimension of the embeddings for categorical features
        (default: [min(50, (cat+1)//2) for cat in cardinality])
    scaling: Whether to automatically scale the target values (default: true)
    lags_seq: Indices of the lagged target values to use as inputs of the RNN
        (default: None, in which case these are automatically determined
        based on freq)
    time_features: Time features to use as inputs of the RNN (default: None, in which
        case these are automatically determined based on freq)
    num_parallel_samples: Number of evaluation samples per time series to increase parallelism during inference.
        This is a model optimization that does not affect the accuracy (default: 100)
    forecast_type: The model outputs M and Q. This determines how those parameters are translated to a
        regular timeseries forecast.
        flat --> A flat forecast of M/Q for the prediction length
        exact --> Will use M and Q to produce a foreast with (Q) times 0,M, repeat
        hybrid --> Will use M and Q to create flat forecasts, but the M/Q changes depending
        on the subsequent predictions
        eg: We trained the model with prediction length = 5. the output of the model are
        M --> 22,33,12, Q--> 2,1,4
        flat --> 11, 11, 11, 11, 11
        exact --> 0, 22, 33, 0, 0
        hybrid --> 11, 11, 33, 3, 3 (22/2, 22/2, 33/1, 12/4, 12/4)
"""


class DeepRenewalEstimator(GluonEstimator):
    """Construct a DeepRenewal estimator.

    This implements an RNN-based model, close to the one described in
    https://arxiv.org/abs/1911.10416.

    Args:
        freq (str):
            Frequency of the data to train on and predict
        prediction_length (int):
            Length of the prediction horizon
        trainer (Optional[Trainer], optional):
            Trainer object to be used. Defaults to Trainer().
        context_length (Optional[int], optional):
            Number of steps to unroll the RNN for before
            computing predictions. Defaults to None in which case context_length = prediction_length.
        num_layers (Optional[int], optional):
            Number of RNN layers. Defaults to 2.
        num_cells (Optional[int], optional):
            Number of RNN cells for each layer. Defaults to 40.
        cell_type (Optional[str], optional):
            Type of recurrent cells to use (available: 'lstm' or 'gru'). Defaults to "lstm".
        dropout_rate (Optional[float], optional):
            Dropout regularization parameter. Defaults to 0.1.
        use_feat_dynamic_real (Optional[bool], optional):
            Whether to use the ``feat_dynamic_real`` field from the data. Defaults to False.
        use_feat_static_cat (Optional[bool], optional):
            Whether to use the ``feat_static_cat`` field from the data. Defaults to False.
        use_feat_static_real (Optional[bool], optional):
            Whether to use the ``feat_static_real`` field from the data. Defaults to False.
        cardinality (Optional[List[int]], optional):
            Number of values of each categorical feature.
            This must be set if ``use_feat_static_cat == True``. Defaults to None.
        embedding_dimension (Optional[List[int]], optional):
            Dimension of the embeddings for categorical features.
            Defaults to None in which case [min(50, (cat+1)//2) for cat in cardinality].
        scaling (Optional[bool], optional):
            Whether to automatically scale the target values. Defaults to True.
        lags_seq (Optional[List[int]], optional):
            Indices of the lagged target values to use as inputs of the RNN
            Defaults to None in which case these are automatically determined based on freq.
        time_features (Optional[List[TimeFeature]], optional):
            Time features to use as inputs of the RNN.
            Defaults to None in which case these are automatically determined based on freq.
        num_parallel_samples (Optional[int], optional):
            Number of evaluation samples per time series to increase parallelism during inference.
            This is a model optimization that does not affect the accuracy. Defaults to 100.
        forecast_type (Optional[str], optional):
            The model outputs M and Q. This determines how those parameters are translated to a
            regular timeseries forecast.
            flat --> A flat forecast of M/Q for the prediction length
            exact --> Will use M and Q to produce a foreast with (Q) times 0,M, repeat
            hybrid --> Will use M and Q to create flat forecasts, but the M/Q changes depending
            on the subsequent predictions
            eg: We trained the model with prediction length = 5. the output of the model are
            M --> 22,33,12, Q--> 2,1,4
            flat --> 11, 11, 11, 11, 11
            exact --> 0, 22, 33, 0, 0
            hybrid --> 11, 11, 33, 3, 3 (22/2, 22/2, 33/1, 12/4, 12/4). Defaults to "flat".
    """

    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        trainer: Optional[Trainer] = Trainer(),
        context_length: Optional[int] = None,
        num_layers: Optional[int] = 2,
        num_cells: Optional[int] = 40,
        cell_type: Optional[str] = "lstm",
        dropout_rate: Optional[float] = 0.1,
        use_feat_dynamic_real: Optional[bool] = False,
        use_feat_static_cat: Optional[bool] = False,
        use_feat_static_real: Optional[bool] = False,
        cardinality: Optional[List[int]] = None,
        embedding_dimension: Optional[List[int]] = None,
        scaling: Optional[bool] = True,
        lags_seq: Optional[List[int]] = None,
        time_features: Optional[List[TimeFeature]] = None,
        num_parallel_samples: Optional[int] = 100,
        forecast_type: Optional[str] = "flat",
        dtype: Optional[DType] = np.float32,
    ) -> None:

        super().__init__(trainer=trainer, dtype=dtype)

        assert prediction_length > 0, "The value of `prediction_length` should be > 0"
        assert (
            context_length is None or context_length > 0
        ), "The value of `context_length` should be > 0"
        assert num_layers > 0, "The value of `num_layers` should be > 0"
        assert num_cells > 0, "The value of `num_cells` should be > 0"
        assert dropout_rate >= 0, "The value of `dropout_rate` should be >= 0"
        assert (cardinality is not None and use_feat_static_cat) or (
            cardinality is None and not use_feat_static_cat
        ), "You should set `cardinality` if and only if `use_feat_static_cat=True`"
        assert cardinality is None or all(
            [c > 0 for c in cardinality]
        ), "Elements of `cardinality` should be > 0"
        assert embedding_dimension is None or all(
            [e > 0 for e in embedding_dimension]
        ), "Elements of `embedding_dimension` should be > 0"
        assert (
            num_parallel_samples > 0
        ), "The value of `num_parallel_samples` should be > 0"

        self.freq = freq
        self.context_length = (
            context_length if context_length is not None else prediction_length
        )
        self.prediction_length = prediction_length
        self.distr_output_m = NegativeBinomialOutput()
        self.distr_output_q = NegativeBinomialOutput()
        self.distr_output_m.dtype = dtype
        self.distr_output_q.dtype = dtype
        self.num_layers = num_layers
        self.num_cells = num_cells
        self.cell_type = cell_type
        self.dropout_rate = dropout_rate
        self.use_feat_dynamic_real = use_feat_dynamic_real
        self.use_feat_static_cat = use_feat_static_cat
        self.use_feat_static_real = use_feat_static_real
        self.cardinality = cardinality if cardinality and use_feat_static_cat else [1]
        self.embedding_dimension = (
            embedding_dimension
            if embedding_dimension is not None
            else [min(50, (cat + 1) // 2) for cat in self.cardinality]
        )
        self.scaling = scaling
        self.lags_seq = (
            lags_seq if lags_seq is not None else get_lags_for_frequency(freq_str=freq)
        )
        self.time_features = (
            time_features
            if time_features is not None
            else time_features_from_frequency_str(self.freq)
        )

        self.history_length = self.context_length + max(self.lags_seq)

        self.num_parallel_samples = num_parallel_samples
        self.forecast_type = forecast_type

    def create_transformation(self) -> Transformation:
        remove_field_names = [FieldName.FEAT_DYNAMIC_CAT]
        if not self.use_feat_static_real:
            remove_field_names.append(FieldName.FEAT_STATIC_REAL)
        if not self.use_feat_dynamic_real:
            remove_field_names.append(FieldName.FEAT_DYNAMIC_REAL)

        return Chain(
            [RemoveFields(field_names=remove_field_names)]
            + (
                [SetField(output_field=FieldName.FEAT_STATIC_CAT, value=[0.0])]
                if not self.use_feat_static_cat
                else []
            )
            + (
                [SetField(output_field=FieldName.FEAT_STATIC_REAL, value=[0.0])]
                if not self.use_feat_static_real
                else []
            )
            + [
                AsNumpyArray(
                    field=FieldName.FEAT_STATIC_CAT,
                    expected_ndim=1,
                    dtype=self.dtype,
                ),
                AsNumpyArray(
                    field=FieldName.FEAT_STATIC_REAL,
                    expected_ndim=1,
                    dtype=self.dtype,
                ),
                AsNumpyArray(
                    field=FieldName.TARGET,
                    # in the following line, we add 1 for the time dimension
                    expected_ndim=1 + len(self.distr_output_m.event_shape),
                    dtype=self.dtype,
                ),
                AddObservedValuesIndicator(
                    target_field=FieldName.TARGET,
                    output_field=FieldName.OBSERVED_VALUES,
                    dtype=self.dtype,
                ),
                AddTimeFeatures(
                    start_field=FieldName.START,
                    target_field=FieldName.TARGET,
                    output_field=FieldName.FEAT_TIME,
                    time_features=self.time_features,
                    pred_length=self.prediction_length,
                ),
                AddInterDemandPeriodFeature(
                    start_field=FieldName.START,
                    target_field=FieldName.TARGET,
                    output_field=FieldName.TARGET,  # FieldName.FEAT_TIME FieldName.TARGET #if we want to append to feat time,
                    pred_length=self.prediction_length,
                ),
                AddAgeFeature(
                    target_field=FieldName.TARGET,
                    output_field=FieldName.FEAT_AGE,
                    pred_length=self.prediction_length,
                    log_scale=True,
                    dtype=self.dtype,
                ),
                VstackFeatures(
                    output_field=FieldName.FEAT_TIME,
                    input_fields=[FieldName.FEAT_TIME, FieldName.FEAT_AGE]
                    + (
                        [FieldName.FEAT_DYNAMIC_REAL]
                        if self.use_feat_dynamic_real
                        else []
                    ),
                ),
                # DropNonZeroTarget(
                #     input_fields=[FieldName.FEAT_TIME, FieldName.OBSERVED_VALUES],
                #     target_field=FieldName.TARGET,
                #     pred_length=self.prediction_length,
                # ),
                RenewalInstanceSplitter(
                    target_field=FieldName.TARGET,
                    is_pad_field=FieldName.IS_PAD,
                    start_field=FieldName.START,
                    forecast_start_field=FieldName.FORECAST_START,
                    train_sampler=ExpectedNumInstanceSampler(num_instances=1),
                    past_length=self.history_length,
                    future_length=self.prediction_length,
                    time_series_fields=[FieldName.FEAT_TIME, FieldName.OBSERVED_VALUES],
                    # dummy_value=self.distr_output_m.value_in_support,
                    # pick_incomplete=False
                ),
            ]
        )

    def create_training_network(self) -> DeepRenewalTrainingNetwork:
        return DeepRenewalTrainingNetwork(
            num_layers=self.num_layers,
            num_cells=self.num_cells,
            cell_type=self.cell_type,
            history_length=self.history_length,
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            distr_output_m=self.distr_output_m,
            distr_output_q=self.distr_output_q,
            dropout_rate=self.dropout_rate,
            cardinality=self.cardinality,
            embedding_dimension=self.embedding_dimension,
            lags_seq=self.lags_seq,
            scaling=self.scaling,
            dtype=self.dtype,
        )

    def create_predictor(
        self, transformation: Transformation, trained_network: HybridBlock
    ) -> Predictor:
        prediction_network = DeepRenewalPredictionNetwork(
            num_parallel_samples=self.num_parallel_samples,
            num_layers=self.num_layers,
            num_cells=self.num_cells,
            cell_type=self.cell_type,
            history_length=self.history_length,
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            distr_output_m=self.distr_output_m,
            distr_output_q=self.distr_output_q,
            dropout_rate=self.dropout_rate,
            cardinality=self.cardinality,
            embedding_dimension=self.embedding_dimension,
            lags_seq=self.lags_seq,
            scaling=self.scaling,
            dtype=self.dtype,
        )

        copy_parameters(trained_network, prediction_network)

        return RepresentableBlockPredictor(
            input_transform=transformation,
            prediction_net=prediction_network,
            batch_size=self.trainer.batch_size,
            freq=self.freq,
            prediction_length=self.prediction_length,
            ctx=self.trainer.ctx,
            dtype=self.dtype,
            forecast_generator=IntermittentSampleForecastGenerator(
                prediction_length=self.prediction_length,
                forecast_type=self.forecast_type,
            ),
        )



# Standard library imports
import logging
import os
from typing import Dict, Iterator, Optional

# Third-party imports
import numpy as np
import pandas as pd

# First-party imports
from gluonts.core.component import validated
from gluonts.dataset.common import Dataset
from gluonts.model.forecast import SampleForecast
from gluonts.model.predictor import RepresentablePredictor
from gluonts.support.pandas import forecast_start

from .croston import fit_croston

logger = logging.getLogger("gluonts").getChild("croston")


class CrostonForecastPredictor(RepresentablePredictor):
    """
    Wrapper for calling the `croston forecast

    Parameters
    ----------
    freq
        The granularity of the time series (e.g. '1H')
    prediction_length
        Number of time points to be predicted.
    variant
        The method from rforecast to be used one of
        "original", "sba", "sbj"
    trunc_length
        Maximum history length to feed to the model (some models become slow
        with very long series).
    """

    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        variant: str = "original",
        no_of_params: int = 2,
        trunc_length: Optional[int] = None,
    ) -> None:
        super().__init__(freq=freq, prediction_length=prediction_length)

        supported_methods = ["original", "sba", "sbj"]
        assert (
            variant in supported_methods
        ), f"method {variant} is not supported please use one of {supported_methods}"
        self.variant = variant
        assert (
            no_of_params <= 2,
            "Number of parameters should be less than or equal 2",
        )
        self.no_of_params = no_of_params

        self.prediction_length = prediction_length
        self.freq = freq
        self.trunc_length = trunc_length
        self.params = {
            "prediction_length": self.prediction_length,
            "output_types": ["samples"],
        }

    def _unlist(self, l):
        if type(l).__name__.endswith("Vector"):
            return [self._unlist(x) for x in l]
        else:
            return l

    def _run_croston_forecast(self, d, params):
        fit_pred = fit_croston(
            d["target"],
            forecast_length=params["prediction_length"],
            croston_variant=self.variant,
            number_parameters=self.no_of_params,
        )

        return {"samples": fit_pred["croston_forecast"].reshape(1, -1)}

    def predict(
        self,
        dataset: Dataset,
        num_samples: int = 1,
        save_info: bool = False,
        **kwargs,
    ) -> Iterator[SampleForecast]:
        if num_samples != 1:
            num_samples = 1
            logger.warning(
                "num_samples changed to 1 because Croston is non-probabilistic"
            )

        assert num_samples == 1, "Non Probabilistic Method only supports num_samples=1"

        for entry in dataset:
            if isinstance(entry, dict):
                data = entry
            else:
                data = entry.data
                if self.trunc_length:
                    data = data[-self.trunc_length :]

            params = self.params.copy()
            params["num_samples"] = num_samples
            forecast_dict = self._run_croston_forecast(data, params)
            samples = np.array(forecast_dict["samples"])
            expected_shape = (params["num_samples"], self.prediction_length)
            assert (
                samples.shape == expected_shape
            ), f"Expected shape {expected_shape} but found {samples.shape}"

            yield SampleForecast(
                samples, forecast_start(data), self.freq, item_id=data["item_id"]
            )



import logging
from typing import Any, Callable, Iterator, List, Optional

# Third-party imports
import mxnet as mx
import numpy as np

# First-party imports
from gluonts.core.component import validated
from gluonts.dataset.common import DataEntry
from gluonts.dataset.field_names import FieldName
from gluonts.dataset.loader import InferenceDataLoader
from gluonts.model.forecast import (
    DistributionForecast,
    Forecast,
    QuantileForecast,
    SampleForecast,
)
from gluonts.model.forecast_generator import ForecastGenerator

OutputTransform = Callable[[DataEntry, np.ndarray], np.ndarray]
BlockType = mx.gluon.Block

ALLOWABLE_FORECAST_TYPES = ["flat", "exact", "hybrid"]


class IntermittentSampleForecastGenerator(ForecastGenerator):
    @validated()
    def __init__(self, prediction_length: int, forecast_type: str):
        self.prediction_length = prediction_length
        self.forecast_type = forecast_type
        assert (
            forecast_type in ALLOWABLE_FORECAST_TYPES
        ), "Forecast Type needs to be one of " + "|".join(ALLOWABLE_FORECAST_TYPES)

    def __call__(
        self,
        inference_data_loader: InferenceDataLoader,
        prediction_net: BlockType,
        input_names: List[str],
        freq: str,
        output_transform: Optional[OutputTransform],
        num_samples: Optional[int],
        **kwargs,
    ) -> Iterator[Forecast]:
        for batch in inference_data_loader:
            inputs = [batch[k] for k in input_names]
            outputs = prediction_net(*inputs).asnumpy()
            if output_transform is not None:
                outputs = output_transform(batch, outputs)
            if num_samples:
                num_collected_samples = outputs[0].shape[0]
                collected_samples = [outputs]
                while num_collected_samples < num_samples:
                    outputs = prediction_net(*inputs).asnumpy()
                    if output_transform is not None:
                        outputs = output_transform(batch, outputs)
                    collected_samples.append(outputs)
                    num_collected_samples += outputs[0].shape[0]
                outputs = [
                    np.concatenate(s)[:num_samples] for s in zip(*collected_samples)
                ]
                assert len(outputs[0]) == num_samples
            i = -1

            for i, output in enumerate(outputs):
                # M/ Q
                if self.forecast_type == "flat":
                    output = output[:, :, 0] / output[:, :, 1]
                    mask_nan = np.isnan(output)
                    mask_inf = np.isposinf(output)
                    output[mask_nan] = 0
                    output[mask_inf] = 0
                    output = np.vstack([output[:, 0]] * self.prediction_length).T
                # exact: --> (Q-1) times 0, M, (Q-1) times 0, M, repeat
                # hybrid: --> Q times M/Q, Q times M/Q, repeat
                elif self.forecast_type in ["exact", "hybrid"]:
                    output_list = []
                    for o in output:
                        pred = []
                        for row in o:
                            m = row[0]
                            q = max(1, row[1])
                            if self.forecast_type == "hybrid":
                                f = m / q if m != 0 else 0
                            else:
                                f = 0
                            for j in range(int(q - 1)):
                                pred.append(f)
                            pred.append(
                                m
                            ) if self.forecast_type == "exact" else pred.append(f)
                        output_list.append(pred[: self.prediction_length])
                    output = np.array(output_list)
                else:
                    raise NotImplementedError(
                        f"{self.forecast_type} is not a value choice for forecast_type"
                    )
                yield SampleForecast(
                    output,
                    start_date=batch["forecast_start"][i],
                    freq=freq,
                    item_id=batch[FieldName.ITEM_ID][i]
                    if FieldName.ITEM_ID in batch
                    else None,
                    info=batch["info"][i] if "info" in batch else None,
                )
            assert i + 1 == len(batch["forecast_start"])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from copy import deepcopy
from typing import List, Optional, Tuple

# Third-party imports
import mxnet as mx

# Standard library imports
import numpy as np

# First-party imports
from gluonts.block.feature import FeatureEmbedder
from gluonts.block.scaler import MeanScaler, NOPScaler

# TODO bring back validation once model is ready
from gluonts.core.component import DType, validated
from gluonts.distribution import Distribution, DistributionOutput
from gluonts.distribution.distribution import getF
from gluonts.model.common import Tensor
from gluonts.support.util import weighted_average


def prod(xs):
    p = 1
    for x in xs:
        p *= x
    return p


class DeepRenewalNetwork(mx.gluon.HybridBlock):
    @validated()
    def __init__(
        self,
        num_layers: int,
        num_cells: int,
        cell_type: str,
        history_length: int,
        context_length: int,
        prediction_length: int,
        distr_output_m: DistributionOutput,
        distr_output_q: DistributionOutput,
        dropout_rate: float,
        cardinality: List[int],
        embedding_dimension: List[int],
        lags_seq: List[int],
        scaling: bool = True,
        dtype: DType = np.float32,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.num_layers = num_layers
        self.num_cells = num_cells
        self.cell_type = cell_type
        self.history_length = history_length
        self.context_length = context_length
        self.prediction_length = prediction_length
        self.dropout_rate = dropout_rate
        self.cardinality = cardinality
        self.embedding_dimension = embedding_dimension
        self.num_cat = len(cardinality)
        self.scaling = scaling
        self.dtype = dtype

        assert len(cardinality) == len(
            embedding_dimension
        ), "embedding_dimension should be a list with the same size as cardinality"

        assert len(set(lags_seq)) == len(lags_seq), "no duplicated lags allowed!"
        lags_seq.sort()

        self.lags_seq = lags_seq
        # 2 separate distributions
        self.distr_output_m = distr_output_m
        self.distr_output_q = distr_output_q
        RnnCell = {"lstm": mx.gluon.rnn.LSTMCell, "gru": mx.gluon.rnn.GRUCell}[
            self.cell_type
        ]

        self.target_shape = distr_output_m.event_shape

        # TODO: is the following restriction needed?
        assert (
            len(self.target_shape) <= 1
        ), "Argument `target_shape` should be a tuple with 1 element at most"

        with self.name_scope():
            self.proj_distr_args_m = self.distr_output_m.get_args_proj(prefix="m")
            self.proj_distr_args_q = self.distr_output_q.get_args_proj(prefix="q")
            self.rnn = mx.gluon.rnn.HybridSequentialRNNCell()
            for k in range(num_layers):
                cell = RnnCell(hidden_size=num_cells)
                cell = mx.gluon.rnn.ResidualCell(cell) if k > 0 else cell
                cell = (
                    mx.gluon.rnn.ZoneoutCell(cell, zoneout_states=dropout_rate)
                    if dropout_rate > 0.0
                    else cell
                )
                self.rnn.add(cell)
            self.rnn.cast(dtype=dtype)
            self.embedder = FeatureEmbedder(
                cardinalities=cardinality,
                embedding_dims=embedding_dimension,
                dtype=self.dtype,
            )
            if scaling:
                self.scaler = MeanScaler(keepdims=True)
            else:
                self.scaler = NOPScaler(keepdims=True)

    @staticmethod
    def get_lagged_subsequences(
        F,
        sequence: Tensor,
        sequence_length: int,
        indices: List[int],
        subsequences_length: int = 1,
    ) -> Tensor:
        """
        Returns lagged subsequences of a given sequence.
        Parameters
        ----------
        sequence : Tensor
            the sequence from which lagged subsequences should be extracted.
            Shape: (N, T, C).
        sequence_length : int
            length of sequence in the T (time) dimension (axis = 1).
        indices : List[int]
            list of lag indices to be used.
        subsequences_length : int
            length of the subsequences to be extracted.
        Returns
        --------
        lagged : Tensor
            a tensor of shape (N, S, C, I), where S = subsequences_length and
            I = len(indices), containing lagged subsequences. Specifically,
            lagged[i, j, :, k] = sequence[i, -indices[k]-S+j, :].
        """
        # we must have: sequence_length - lag_index - subsequences_length >= 0
        # for all lag_index, hence the following assert
        assert max(indices) + subsequences_length <= sequence_length, (
            f"lags cannot go further than history length, "
            f"found lag {max(indices)} while history length is only "
            f"{sequence_length}"
        )
        assert all(lag_index >= 0 for lag_index in indices)

        lagged_values = []
        for lag_index in indices:
            begin_index = -lag_index - subsequences_length
            end_index = -lag_index if lag_index > 0 else None
            lagged_values.append(
                F.slice_axis(sequence, axis=1, begin=begin_index, end=end_index)
            )

        return F.stack(*lagged_values, axis=-1)

    def unroll_encoder(
        self,
        F,
        feat_static_cat: Tensor,  # (batch_size, num_features)
        feat_static_real: Tensor,  # (batch_size, num_features)
        past_time_feat: Tensor,  # (batch_size, history_length, num_features)
        past_target: Tensor,  # (batch_size, history_length, *target_shape)
        past_observed_values: Tensor,  # (batch_size, history_length, *target_shape)
        future_time_feat: Optional[
            Tensor
        ],  # (batch_size, prediction_length, num_features)
        future_target: Optional[
            Tensor
        ],  # (batch_size, prediction_length, *target_shape)
    ) -> Tuple[Tensor, List, Tensor, Tensor]:
        """
        Unrolls the LSTM encoder over past and, if present, future data.
        Returns outputs and state of the encoder, plus the scale of past_target
        and a vector of static features that was constructed and fed as input
        to the encoder.
        All tensor arguments should have NTC layout.
        """

        if future_time_feat is None or future_target is None:
            time_feat = past_time_feat.slice_axis(
                axis=1,
                begin=self.history_length - self.context_length,
                end=None,
            )
            sequence = past_target
            sequence_length = self.history_length
            subsequences_length = self.context_length
        else:
            time_feat = F.concat(
                past_time_feat.slice_axis(
                    axis=1,
                    begin=self.history_length - self.context_length,
                    end=None,
                ),
                future_time_feat,
                dim=1,
            )
            sequence = F.concat(past_target, future_target, dim=1)
            sequence_length = self.history_length + self.prediction_length
            subsequences_length = self.context_length + self.prediction_length

        # (batch_size, sub_seq_len, *target_shape, num_lags)
        lags_m = self.get_lagged_subsequences(
            F=F,
            sequence=sequence.slice_axis(axis=2, begin=0, end=1).squeeze(
                axis=2
            ),  # sequence[:,:,0]
            sequence_length=sequence_length,
            indices=self.lags_seq,
            subsequences_length=subsequences_length,
        )

        lags_q = self.get_lagged_subsequences(
            F=F,
            sequence=sequence.slice_axis(axis=2, begin=1, end=2).squeeze(
                axis=2
            ),  # sequence[:,:,1],
            sequence_length=sequence_length,
            indices=self.lags_seq,
            subsequences_length=subsequences_length,
        )

        # scale is computed on the context length last units of the past target
        # scale shape is (batch_size, 1, *target_shape)
        # past_target[:,:,0]
        _, scale = self.scaler(
            past_target.slice_axis(axis=2, begin=0, end=1)
            .squeeze(axis=2)
            .slice_axis(axis=1, begin=-self.context_length, end=None),
            past_observed_values.slice_axis(
                axis=1, begin=-self.context_length, end=None
            ),
        )

        # _, scale_q = self.scaler(
        #     past_target[:,:,1].slice_axis(axis=1, begin=-self.context_length, end=None),
        #     past_observed_values.slice_axis(
        #         axis=1, begin=-self.context_length, end=None
        #     ),
        # )

        # (batch_size, num_features)
        embedded_cat = self.embedder(feat_static_cat)

        # in addition to embedding features, use the log scale as it can help
        # prediction too
        # (batch_size, num_features + prod(target_shape))
        static_feat = F.concat(
            embedded_cat,
            feat_static_real,
            F.log(scale)
            if len(self.target_shape) == 0
            else F.log(scale.squeeze(axis=1)),
            dim=1,
        )

        # (batch_size, subsequences_length, num_features + 1)
        repeated_static_feat = static_feat.expand_dims(axis=1).repeat(
            axis=1, repeats=subsequences_length
        )

        # (batch_size, sub_seq_len, *target_shape, num_lags)
        lags_scaled_m = F.broadcast_div(lags_m, scale.expand_dims(axis=-1))
        lags_scaled_q = F.broadcast_div(lags_q, scale.expand_dims(axis=-1))

        # from (batch_size, sub_seq_len, *target_shape, num_lags)
        # to (batch_size, sub_seq_len, prod(target_shape) * num_lags)
        input_lags_m = F.reshape(
            data=lags_scaled_m,
            shape=(
                -1,
                subsequences_length,
                len(self.lags_seq) * prod(self.target_shape),
            ),
        )
        input_lags_q = F.reshape(
            data=lags_scaled_q,
            shape=(
                -1,
                subsequences_length,
                len(self.lags_seq) * prod(self.target_shape),
            ),
        )

        # (batch_size, sub_seq_len, input_dim)
        inputs = F.concat(
            input_lags_m, input_lags_q, time_feat, repeated_static_feat, dim=-1
        )

        # unroll encoder
        outputs, state = self.rnn.unroll(
            inputs=inputs,
            length=subsequences_length,
            layout="NTC",
            merge_outputs=True,
            begin_state=self.rnn.begin_state(
                func=F.zeros,
                dtype=self.dtype,
                batch_size=inputs.shape[0] if isinstance(inputs, mx.nd.NDArray) else 0,
            ),
        )

        # outputs: (batch_size, seq_len, num_cells)
        # state: list of (batch_size, num_cells) tensors
        # scale: (batch_size, 1, *target_shape)
        # static_feat: (batch_size, num_features + prod(target_shape))
        return outputs, state, scale, static_feat


class DeepRenewalTrainingNetwork(DeepRenewalNetwork):
    def distribution(
        self,
        feat_static_cat: Tensor,
        feat_static_real: Tensor,
        past_time_feat: Tensor,
        past_target: Tensor,
        past_observed_values: Tensor,
        future_time_feat: Tensor,
        future_target: Tensor,
        future_observed_values: Tensor,
    ) -> Distribution:
        """

        Returns the distribution predicted by the model on the range of
        past_target and future_target.

        The distribution is obtained by unrolling the network with the true
        target, this is also the distribution that is being minimized during
        training. This can be used in anomaly detection, see for instance
        examples/anomaly_detection.py.

        Input arguments are the same as for the hybrid_forward method.

        Returns
        -------
        Distribution
            a distribution object whose mean has shape:
            (batch_size, context_length + prediction_length).
        """
        # unroll the decoder in "training mode"
        # i.e. by providing future data as well
        F = getF(feat_static_cat)

        rnn_outputs, _, scale, _ = self.unroll_encoder(
            F=F,
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            past_time_feat=past_time_feat,
            past_target=past_target,
            past_observed_values=past_observed_values,
            future_time_feat=future_time_feat,
            future_target=future_target,
        )
        distr_args_m = self.proj_distr_args_m(rnn_outputs)
        distr_args_q = self.proj_distr_args_q(rnn_outputs)
        return (
            self.distr_output_m.distribution(distr_args_m, scale=scale),
            self.distr_output_q.distribution(distr_args_q, scale=scale),
        )

    # noinspection PyMethodOverriding,PyPep8Naming
    def hybrid_forward(
        self,
        F,
        feat_static_cat: Tensor,
        feat_static_real: Tensor,
        past_time_feat: Tensor,
        past_target: Tensor,
        past_observed_values: Tensor,
        future_time_feat: Tensor,
        future_target: Tensor,
        future_observed_values: Tensor,
    ) -> Tensor:
        """
        Computes the loss for training DeepRenewalProcess, all inputs tensors representing
        time series have NTC layout.

        Parameters
        ----------
        F
        feat_static_cat : (batch_size, num_features)
        feat_static_real : (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape, seq_len)
        future_time_feat : (batch_size, prediction_length, num_features)
        future_target : (batch_size, prediction_length, *target_shape)
        future_observed_values : (batch_size, prediction_length, *target_shape)

        Returns loss with shape (batch_size, context + prediction_length, 1)
        -------

        """
        distr_m, distr_q = self.distribution(
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            past_time_feat=past_time_feat,
            past_target=past_target,
            past_observed_values=past_observed_values,
            future_time_feat=future_time_feat,
            future_target=future_target,
            future_observed_values=future_observed_values,
        )

        # put together target sequence
        # (batch_size, seq_len, *target_shape)
        target = F.concat(
            past_target.slice_axis(
                axis=1,
                begin=self.history_length - self.context_length,
                end=None,
            ),
            future_target,
            dim=1,
        )

        # (batch_size, seq_len)
        loss_m = distr_m.loss(
            target.slice_axis(axis=2, begin=0, end=1).squeeze()
        )  # target[:,:,0]
        loss_q = distr_q.loss(
            target.slice_axis(axis=2, begin=1, end=2).squeeze()
        )  # target[:,:,1]
        loss = loss_m + loss_q

        # (batch_size, seq_len, *target_shape)
        observed_values = F.concat(
            past_observed_values.slice_axis(
                axis=1,
                begin=self.history_length - self.context_length,
                end=self.history_length,
            ),
            future_observed_values,
            dim=1,
        )

        # mask the loss at one time step iff one or more observations is missing in the target dimensions
        # (batch_size, seq_len)
        loss_weights = (
            observed_values
            if (len(self.target_shape) == 0)
            else observed_values.min(axis=-1, keepdims=False)
        )

        weighted_loss = weighted_average(F=F, x=loss, weights=loss_weights, axis=1)

        # need to mask possible nans and -inf
        loss = F.where(condition=loss_weights, x=loss, y=F.zeros_like(loss))

        return weighted_loss, loss


class DeepRenewalPredictionNetwork(DeepRenewalNetwork):
    @validated()
    def __init__(self, num_parallel_samples: int = 100, **kwargs) -> None:
        super().__init__(**kwargs)
        self.num_parallel_samples = num_parallel_samples

        # for decoding the lags are shifted by one, at the first time-step
        # of the decoder a lag of one corresponds to the last target value
        self.shifted_lags = [l - 1 for l in self.lags_seq]

    def sampling_decoder(
        self,
        F,
        static_feat: Tensor,
        past_target: Tensor,
        time_feat: Tensor,
        scale: Tensor,
        begin_states: List,
    ) -> Tensor:
        """
        Computes sample paths by unrolling the LSTM starting with a initial
        input and state.

        Parameters
        ----------
        static_feat : Tensor
            static features. Shape: (batch_size, num_static_features).
        past_target : Tensor
            target history. Shape: (batch_size, history_length).
        time_feat : Tensor
            time features. Shape: (batch_size, prediction_length, num_time_features).
        scale : Tensor
            tensor containing the scale of each element in the batch. Shape: (batch_size, 1, 1).
        begin_states : List
            list of initial states for the LSTM layers.
            the shape of each tensor of the list should be (batch_size, num_cells)
        Returns
        --------
        Tensor
            A tensor containing sampled paths.
            Shape: (batch_size, num_sample_paths, prediction_length).
        """

        # blows-up the dimension of each tensor to batch_size * self.num_parallel_samples for increasing parallelism
        repeated_past_target = past_target.repeat(
            repeats=self.num_parallel_samples, axis=0
        )
        repeated_time_feat = time_feat.repeat(repeats=self.num_parallel_samples, axis=0)
        repeated_static_feat = static_feat.repeat(
            repeats=self.num_parallel_samples, axis=0
        ).expand_dims(axis=1)
        repeated_scale = scale.repeat(repeats=self.num_parallel_samples, axis=0)
        repeated_states = [
            s.repeat(repeats=self.num_parallel_samples, axis=0) for s in begin_states
        ]

        future_samples = []
        batch_size = past_target.shape[0]
        # for each future time-units we draw new samples for this time-unit and update the state
        for k in range(self.prediction_length):
            # (batch_size * num_samples, 1, *target_shape, num_lags)
            lags_m = self.get_lagged_subsequences(
                F=F,
                sequence=repeated_past_target.slice_axis(
                    axis=2, begin=0, end=1
                ).squeeze(),  # repeated_past_target[:,:,0]
                sequence_length=self.history_length + k,
                indices=self.shifted_lags,
                subsequences_length=1,
            )
            lags_q = self.get_lagged_subsequences(
                F=F,
                sequence=repeated_past_target.slice_axis(
                    axis=2, begin=1, end=2
                ).squeeze(),  # repeated_past_target[:,:,1]
                sequence_length=self.history_length + k,
                indices=self.shifted_lags,
                subsequences_length=1,
            )

            # (batch_size * num_samples, 1, *target_shape, num_lags)
            lags_scaled_m = F.broadcast_div(lags_m, repeated_scale.expand_dims(axis=-1))
            lags_scaled_q = F.broadcast_div(lags_q, repeated_scale.expand_dims(axis=-1))

            # from (batch_size * num_samples, 1, *target_shape, num_lags)
            # to (batch_size * num_samples, 1, prod(target_shape) * num_lags)
            input_lags_m = F.reshape(
                data=lags_scaled_m,
                shape=(-1, 1, prod(self.target_shape) * len(self.lags_seq)),
            )
            input_lags_q = F.reshape(
                data=lags_scaled_q,
                shape=(-1, 1, prod(self.target_shape) * len(self.lags_seq)),
            )

            # (batch_size * num_samples, 1, prod(target_shape) * num_lags + num_time_features + num_static_features)
            decoder_input = F.concat(
                input_lags_m,
                input_lags_q,
                repeated_time_feat.slice_axis(axis=1, begin=k, end=k + 1),
                repeated_static_feat,
                dim=-1,
            )

            # output shape: (batch_size * num_samples, 1, num_cells)
            # state shape: (batch_size * num_samples, num_cells)
            rnn_outputs, repeated_states = self.rnn.unroll(
                inputs=decoder_input,
                length=1,
                begin_state=repeated_states,
                layout="NTC",
                merge_outputs=True,
            )
            distr_args_m = self.proj_distr_args_m(rnn_outputs)
            distr_args_q = self.proj_distr_args_q(rnn_outputs)
            # compute likelihood of target given the predicted parameters
            distr_m = self.distr_output_m.distribution(
                distr_args_m, scale=repeated_scale
            )
            distr_q = self.distr_output_q.distribution(
                distr_args_q, scale=repeated_scale
            )
            # (batch_size * num_samples, 1, *target_shape)
            new_samples_m = distr_m.sample(dtype=self.dtype)
            new_samples_q = distr_q.sample(dtype=self.dtype)
            new_samples = F.concat(new_samples_m, new_samples_q, dim=1)
            new_samples = new_samples.expand_dims(axis=1)
            # new_samples = new_samples_m
            # (batch_size * num_samples, seq_len, *target_shape)
            repeated_past_target = F.concat(repeated_past_target, new_samples, dim=1)
            future_samples.append(new_samples)

        # (batch_size * num_samples, prediction_length, *target_shape)
        samples = F.concat(*future_samples, dim=1)
        # (batch_size, num_samples, prediction_length, *target_shape)
        return samples.reshape(
            shape=(
                (batch_size, self.num_parallel_samples)
                + (self.prediction_length,)
                + (samples.shape[-1],)
            )
        )

    # noinspection PyMethodOverriding,PyPep8Naming
    def hybrid_forward(
        self,
        F,
        feat_static_cat: Tensor,  # (batch_size, num_features)
        feat_static_real: Tensor,  # (batch_size, num_features)
        past_time_feat: Tensor,  # (batch_size, history_length, num_features)
        past_target: Tensor,  # (batch_size, history_length, *target_shape)
        past_observed_values: Tensor,  # (batch_size, history_length, *target_shape)
        future_time_feat: Tensor,  # (batch_size, prediction_length, num_features)
    ) -> Tensor:
        """
        Predicts samples, all tensors should have NTC layout.
        Parameters
        ----------
        F
        feat_static_cat : (batch_size, num_features)
        feat_static_real : (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape)
        future_time_feat : (batch_size, prediction_length, num_features)

        Returns
        -------
        Tensor
            Predicted samples
        """

        # unroll the decoder in "prediction mode", i.e. with past data only
        _, state, scale, static_feat = self.unroll_encoder(
            F=F,
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            past_time_feat=past_time_feat,
            past_target=past_target,
            past_observed_values=past_observed_values,
            future_time_feat=None,
            future_target=None,
        )

        return self.sampling_decoder(
            F=F,
            past_target=past_target,
            time_feat=future_time_feat,
            static_feat=static_feat,
            scale=scale,
            begin_states=state,
        )



# First-party imports
from gluonts.core.component import validated
from gluonts.model.estimator import DummyEstimator

# Relative imports
from ._predictor import CrostonForecastPredictor


class CrostonForecastEstimator(DummyEstimator):
    @validated(
        getattr(CrostonForecastPredictor.__init__, "Model")
    )  # Reuse the model Predictor model
    def __init__(self, **kwargs) -> None:
        super().__init__(predictor_cls=CrostonForecastPredictor, **kwargs)



#!/usr/bin/env python

"""The setup script."""
import os
from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('docs/history.md') as history_file:
    history = history_file.read()

# thelibFolder = os.path.dirname(os.path.realpath(__file__))
requirementPath = 'requirements_dev.txt'
requirements = []
if os.path.isfile(requirementPath):
    with open(requirementPath) as f:
        requirements = f.read().splitlines()

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Manu Joseph",
    author_email='manujosephv@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="An implementation of the DeepRenewal Processes in GluonTS",
    entry_points={
        'console_scripts': [
            'deeprenewal=deeprenewal.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='deeprenewal',
    name='deeprenewal',
    packages=find_packages(include=['deeprenewal', 'deeprenewal.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/manujosephv/deeprenewal',
    version='0.3.1',
    zip_safe=False,
)
